@@ -1,4 +1,6 @@
-/* Copyright (C) 2004-2006 MySQL AB
+/*
+   Copyright (C) 2004-2006 MySQL AB
+    All rights reserved. Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "mysql_priv.h"
 #include "event_queue.h"
@@ -60,8 +63,16 @@ extern "C" int event_queue_element_compare_q(void *, uchar *, uchar *);
 
 int event_queue_element_compare_q(void *vptr, uchar* a, uchar *b)
 {
-  my_time_t lhs = ((Event_queue_element *)a)->execute_at;
-  my_time_t rhs = ((Event_queue_element *)b)->execute_at;
+  Event_queue_element *left = (Event_queue_element *)a;
+  Event_queue_element *right = (Event_queue_element *)b;
+  my_time_t lhs = left->execute_at;
+  my_time_t rhs = right->execute_at;
+
+  if (left->status == Event_parse_data::DISABLED)
+    return right->status != Event_parse_data::DISABLED;
+
+  if (right->status == Event_parse_data::DISABLED)
+    return 1;
 
   return (lhs < rhs ? -1 : (lhs > rhs ? 1 : 0));
 }
@@ -190,7 +201,7 @@ Event_queue::create_event(THD *thd, Event_queue_element *new_element,
 
   /* Will do nothing if the event is disabled */
   new_element->compute_next_execution_time();
-  if (new_element->status != Event_queue_element::ENABLED)
+  if (new_element->status != Event_parse_data::ENABLED)
   {
     delete new_element;
     *created= FALSE;
@@ -228,8 +239,8 @@ Event_queue::update_event(THD *thd, LEX_STRING dbname, LEX_STRING name,
   DBUG_ENTER("Event_queue::update_event");
   DBUG_PRINT("enter", ("thd: 0x%lx  et=[%s.%s]", (long) thd, dbname.str, name.str));
 
-  if ((new_element->status == Event_queue_element::DISABLED) ||
-      (new_element->status == Event_queue_element::SLAVESIDE_DISABLED))
+  if ((new_element->status == Event_parse_data::DISABLED) ||
+      (new_element->status == Event_parse_data::SLAVESIDE_DISABLED))
   {
     DBUG_PRINT("info", ("The event is disabled."));
     /*
@@ -434,8 +445,34 @@ Event_queue::recalculate_activation_times(THD *thd)
     ((Event_queue_element*)queue_element(&queue, i))->update_timing_fields(thd);
   }
   queue_fix(&queue);
+  /*
+    The disabled elements are moved to the end during the `fix`.
+    Start from the end and remove all of the elements which are
+    disabled. When we find the first non-disabled one we break, as we
+    have removed all. The queue has been ordered in a way the disabled
+    events are at the end.
+  */
+  for (i= queue.elements; i > 0; i--)
+  {
+    Event_queue_element *element = (Event_queue_element*)queue_element(&queue, i - 1);
+    if (element->status != Event_parse_data::DISABLED)
+      break;
+    /*
+      This won't cause queue re-order, because we remove
+      always the last element.
+    */
+    queue_remove(&queue, i - 1);
+    delete element;
+  }
   UNLOCK_QUEUE_DATA();
 
+  /*
+    XXX: The events are dropped only from memory and not from disk
+         even if `drop_list[j]->dropped` is TRUE. There will be still on the
+         disk till next server restart.
+         Please add code here to do it.
+  */
+
   DBUG_VOID_RETURN;
 }
 
@@ -581,14 +618,14 @@ Event_queue::get_top_for_execution_if_time(THD *thd,
     DBUG_PRINT("info", ("Ready for execution"));
     top->mark_last_executed(thd);
     if (top->compute_next_execution_time())
-      top->status= Event_queue_element::DISABLED;
+      top->status= Event_parse_data::DISABLED;
     DBUG_PRINT("info", ("event %s status is %d", top->name.str, top->status));
 
     top->execution_count++;
     (*event_name)->dropped= top->dropped;
 
     top->update_timing_fields(thd);
-    if (top->status == Event_queue_element::DISABLED)
+    if (top->status == Event_parse_data::DISABLED)
     {
       DBUG_PRINT("info", ("removing from the queue"));
       sql_print_information("Event Scheduler: Last execution of %s.%s. %s",
