@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2007 MySQL AB, 2008, 2009 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <GlobalSignalNumbers.h>
 
@@ -64,7 +67,6 @@ const GsnName SignalNames [] = {
   ,{ GSN_ACC_COMMITREQ,          "ACC_COMMITREQ" }
   ,{ GSN_ACC_OVER_REC,           "ACC_OVER_REC" }
   ,{ GSN_ACC_SCAN_INFO,          "ACC_SCAN_INFO" }
-  ,{ GSN_ACC_SCAN_INFO24,        "ACC_SCAN_INFO24" }
   ,{ GSN_ACC_SCANCONF,           "ACC_SCANCONF" }
   ,{ GSN_ACC_SCANREF,            "ACC_SCANREF" }
   ,{ GSN_ACC_SCANREQ,            "ACC_SCANREQ" }
@@ -256,7 +258,6 @@ const GsnName SignalNames [] = {
   ,{ GSN_NEXT_SCANCONF,          "NEXT_SCANCONF" }
   ,{ GSN_NEXT_SCANREF,           "NEXT_SCANREF" }
   ,{ GSN_NEXT_SCANREQ,           "NEXT_SCANREQ" }
-  ,{ GSN_NEXTOPERATION,          "NEXTOPERATION" }
   ,{ GSN_NF_COMPLETEREP,         "NF_COMPLETEREP" }
   ,{ GSN_OPEN_COMCONF,           "OPEN_COMCONF" }
   ,{ GSN_OPEN_COMREF,            "OPEN_COMREF" }
@@ -646,5 +647,17 @@ const GsnName SignalNames [] = {
   ,{ GSN_PREPARE_COPY_FRAG_CONF,  "PREPARE_COPY_FRAG_CONF" }
 
   ,{ GSN_UPGRADE_PROTOCOL_ORD, "UPGRADE_PROTOCOL_ORD" }
+
+  ,{ GSN_TC_HBREP, "TC_HBREP" }
+  
+  ,{ GSN_START_TOREF, "START_TOREF" }
+  ,{ GSN_END_TOREF, "END_TOREF" }
+  ,{ GSN_START_PERMREP, "START_PERMREP" }
+
+  ,{ GSN_LOCAL_ROUTE_ORD, "LOCAL_ROUTE_ORD" }
+
+  ,{ GSN_ALLOC_MEM_REQ, "ALLOC_MEM_REQ" }
+  ,{ GSN_ALLOC_MEM_REF, "ALLOC_MEM_REF" }
+  ,{ GSN_ALLOC_MEM_CONF, "ALLOC_MEM_CONF" }
 };
 const unsigned short NO_OF_SIGNAL_NAMES = sizeof(SignalNames)/sizeof(GsnName);
