@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "Cmvmi.hpp"
 
@@ -35,6 +37,7 @@
 #include <signaldata/EventSubscribeReq.hpp>
 #include <signaldata/DumpStateOrd.hpp>
 #include <signaldata/DisconnectRep.hpp>
+#include <signaldata/AllocMem.hpp>
 
 #include <EventLogger.hpp>
 #include <TimeQueue.hpp>
@@ -43,7 +46,7 @@
 #include <SafeCounter.hpp>
 
 // Used here only to print event reports on stdout/console.
-EventLogger g_eventLogger;
+extern EventLogger * g_eventLogger;
 extern int simulate_error_during_shutdown;
 
 Cmvmi::Cmvmi(Block_context& ctx) :
@@ -88,7 +91,10 @@ Cmvmi::Cmvmi(Block_context& ctx) :
 
   addRecSignal(GSN_TESTSIG, &Cmvmi::execTESTSIG);
   addRecSignal(GSN_NODE_START_REP, &Cmvmi::execNODE_START_REP, true);
-  
+
+  addRecSignal(GSN_ALLOC_MEM_REF, &Cmvmi::execALLOC_MEM_REF);
+  addRecSignal(GSN_ALLOC_MEM_CONF, &Cmvmi::execALLOC_MEM_CONF);
+
   subscriberPool.setSize(5);
   
   const ndb_mgm_configuration_iterator * db = m_ctx.m_config.getOwnConfigIterator();
@@ -212,6 +218,15 @@ void Cmvmi::execEVENT_REP(Signal* signal)
   if (nodeId == 0)
   {
     nodeId= refToNode(signal->getSendersBlockRef());
+
+    if (nodeId == 0)
+    {
+      /* Event reporter supplied no node id,
+       * assume it was local
+       */
+      nodeId= getOwnNodeId();
+    }
+
     eventReport->setNodeId(nodeId);
   }
 
@@ -241,8 +256,8 @@ void Cmvmi::execEVENT_REP(Signal* signal)
   }
 
   // Print the event info
-  g_eventLogger.log(eventReport->getEventType(), 
-		    signal->theData, signal->getLength(), 0, 0);
+  g_eventLogger->log(eventReport->getEventType(), 
+                     signal->theData, signal->getLength(), 0, 0);
   
   return;
 }//execEVENT_REP()
@@ -356,15 +371,49 @@ Cmvmi::execREAD_CONFIG_REQ(Signal* signal)
 
   Uint32 tupmem = 0;
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_PAGE, &tupmem));
+
+  Uint64 accmem = 0;
+  ndb_mgm_get_int64_parameter(p, CFG_DB_INDEX_MEM, &accmem);
+  if (accmem)
+  {
+    accmem /= GLOBAL_PAGE_SIZE;
+    accmem += (32 / 4); // Added as safty in Configuration.cpp
+    tupmem += accmem;
+  }
+
   if (tupmem)
   {
     Resource_limit rl;
     rl.m_min = tupmem;
     rl.m_max = tupmem;
-    rl.m_resource_id = 3;
+    rl.m_resource_id = RG_DATAMEM;
     m_ctx.m_mm.set_resource_limit(rl);
   }
 
+  {
+    Uint32 redomem = 0;
+    ndb_mgm_get_int_parameter(p, CFG_DB_REDO_BUFFER,
+                              &redomem);
+
+    if (redomem)
+    {
+      redomem /= GLOBAL_PAGE_SIZE;
+      Uint32 tmp = redomem & 15;
+      if (tmp != 0)
+      {
+        redomem += (16 - tmp);
+      }
+
+      Resource_limit rl;
+      rl.m_min = redomem;
+      rl.m_max = redomem;
+      rl.m_resource_id = RG_FILE_BUFFERS;
+      m_ctx.m_mm.set_resource_limit(rl);
+
+      shared_mem += redomem;
+    }
+  }
+
   if (shared_mem+tupmem)
   {
     Resource_limit rl;
@@ -373,40 +422,76 @@ Cmvmi::execREAD_CONFIG_REQ(Signal* signal)
     rl.m_resource_id = 0;
     m_ctx.m_mm.set_resource_limit(rl);
   }
-  
-  if (!m_ctx.m_mm.init())
+
+  // sloppy
+  ndbrequire(refToMain(ref) == NDBCNTR &&
+             (refToNode(ref) == getOwnNodeId() || refToNode(ref) == 0));
   {
-    char buf[255];
-
-    struct ndb_mgm_param_info dm;
-    struct ndb_mgm_param_info sga;
-    size_t size = sizeof(ndb_mgm_param_info);
-    
-    ndb_mgm_get_db_parameter_info(CFG_DB_DATA_MEM, &dm, &size);
-    size = sizeof(ndb_mgm_param_info);
-    ndb_mgm_get_db_parameter_info(CFG_DB_SGA, &sga, &size);
-
-    BaseString::snprintf(buf, sizeof(buf), 
-			 "Malloc (%lld bytes) for %s and %s failed", 
-			 Uint64(shared_mem + tupmem) * 32768,
-			 dm.m_name, sga.m_name);
-    
-    ErrorReporter::handleAssert(buf,
-				__FILE__, __LINE__, NDBD_EXIT_MEMALLOC);
-    
-    ndbrequire(false);
+    AllocMemReq* req = (AllocMemReq*)signal->getDataPtrSend();
+    req->senderRef = reference();
+    req->senderData = senderData;
+    req->requestInfo = AllocMemReq::RT_INIT;
+    sendSignal(NDBFS_REF, GSN_ALLOC_MEM_REQ, signal,
+               AllocMemReq::SignalLength, JBB);
   }
+}
+
+void
+Cmvmi::execALLOC_MEM_CONF(Signal* signal)
+{
+  jamEntry();
+  AllocMemConf * conf = (AllocMemConf*)signal->getDataPtr();
+  Uint32 ref = NDBCNTR_REF; // sloppy
+  Uint32 senderData = conf->senderData;
 
   {
     void* ptr = m_ctx.m_mm.get_memroot();
     m_shared_page_pool.set((GlobalPage*)ptr, ~0);
   }
+
+  {
+    ReadConfigConf * conf = (ReadConfigConf*)signal->getDataPtrSend();
+    conf->senderRef = reference();
+    conf->senderData = senderData;
+    sendSignal(ref, GSN_READ_CONFIG_CONF, signal,
+               ReadConfigConf::SignalLength, JBB);
+  }
+}
+
+void
+Cmvmi::execALLOC_MEM_REF(Signal* signal)
+{
+  jamEntry();
+  char buf[255];
+
+  const ndb_mgm_configuration_iterator * p =
+    m_ctx.m_config.getOwnConfigIterator();
+  ndbrequire(p != 0);
+
+  Uint64 shared_mem = 8*1024*1024;
+  ndb_mgm_get_int64_parameter(p, CFG_DB_SGA, &shared_mem);
+  shared_mem /= GLOBAL_PAGE_SIZE;
+
+  Uint32 tupmem = 0;
+  ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_PAGE, &tupmem));
+
+  struct ndb_mgm_param_info dm;
+  struct ndb_mgm_param_info sga;
+  size_t size = sizeof(ndb_mgm_param_info);
+
+  ndb_mgm_get_db_parameter_info(CFG_DB_DATA_MEM, &dm, &size);
+  size = sizeof(ndb_mgm_param_info);
+  ndb_mgm_get_db_parameter_info(CFG_DB_SGA, &sga, &size);
+
+  BaseString::snprintf(buf, sizeof(buf),
+                       "Malloc (%lld bytes) for %s and %s failed",
+                       Uint64(shared_mem + tupmem) * 32768,
+                       dm.m_name, sga.m_name);
+
+  ErrorReporter::handleAssert(buf,
+                              __FILE__, __LINE__, NDBD_EXIT_MEMALLOC);
   
-  ReadConfigConf * conf = (ReadConfigConf*)signal->getDataPtrSend();
-  conf->senderRef = reference();
-  conf->senderData = senderData;
-  sendSignal(ref, GSN_READ_CONFIG_CONF, signal, 
-	     ReadConfigConf::SignalLength, JBB);
+  ndbrequire(false);
 }
 
 void Cmvmi::execSTTOR(Signal* signal)
@@ -419,13 +504,37 @@ void Cmvmi::execSTTOR(Signal* signal)
 
     if(m_ctx.m_config.lockPagesInMainMemory() == 1)
     {
+      jam();
+      /**
+       * Notify watchdog that we're locking memory...
+       *   which can be equally "heavy" as allocating it
+       */
+      refresh_watch_dog(9);
       int res = NdbMem_MemLockAll(0);
-      if(res != 0){
-	g_eventLogger.warning("Failed to memlock pages");
-	warningEvent("Failed to memlock pages");
+      if (res != 0)
+      {
+        char buf[100];
+        BaseString::snprintf(buf, sizeof(buf), 
+                             "Failed to memlock pages, error: %d (%s)",
+                             errno, strerror(errno));
+        g_eventLogger->warning("%s", buf);
+        warningEvent("%s", buf);
       }
     }
     
+    /**
+     * Install "normal" watchdog value
+     */
+    {
+      Uint32 db_watchdog_interval = 0;
+      const ndb_mgm_configuration_iterator * p = 
+        m_ctx.m_config.getOwnConfigIterator();
+      ndb_mgm_get_int_parameter(p, CFG_DB_WATCHDOG_INTERVAL, 
+                                &db_watchdog_interval);
+      ndbrequire(db_watchdog_interval);
+      update_watch_dog_timer(db_watchdog_interval);
+    }
+    
     sendSTTORRY(signal);
     return;
   } else if (theStartPhase == 3) {
@@ -433,13 +542,26 @@ void Cmvmi::execSTTOR(Signal* signal)
     globalData.activateSendPacked = 1;
     sendSTTORRY(signal);
   } else if (theStartPhase == 8){
-    /*---------------------------------------------------*/
-    /* Open com to API + REP nodes                       */
-    /*---------------------------------------------------*/
-    signal->theData[0] = 0; // no answer
-    signal->theData[1] = 0; // no id
-    signal->theData[2] = NodeInfo::API;
-    execOPEN_COMREQ(signal);
+#ifdef ERROR_INSERT
+    if (ERROR_INSERTED(9004))
+    {
+      Uint32 len = signal->getLength();
+      Uint32 db = c_dbNodes.find(0);
+      if (db == getOwnNodeId())
+        db = c_dbNodes.find(db);
+      Uint32 i = c_error_9000_nodes_mask.find(0);
+      Uint32 tmp[25];
+      memcpy(tmp, signal->theData, sizeof(tmp));
+      signal->theData[0] = i;
+      sendSignal(calcQmgrBlockRef(db),GSN_API_FAILREQ, signal, 1, JBA);
+      ndbout_c("stopping %u using %u", i, db);
+      CLEAR_ERROR_INSERT_VALUE;
+      memcpy(signal->theData, tmp, sizeof(tmp));
+      sendSignalWithDelay(reference(), GSN_STTOR,
+                          signal, 100, len);
+      return;
+    }
+#endif
     globalData.theStartLevel = NodeState::SL_STARTED;
     sendSTTORRY(signal);
   }
@@ -453,6 +575,7 @@ void Cmvmi::execCLOSE_COMREQ(Signal* signal)
   CloseComReqConf * const closeCom = (CloseComReqConf *)&signal->theData[0];
 
   const BlockReference userRef = closeCom->xxxBlockRef;
+  Uint32 requestType = closeCom->requestType;
   Uint32 failNo = closeCom->failNo;
 //  Uint32 noOfNodes = closeCom->noOfNodes;
   
@@ -475,11 +598,20 @@ void Cmvmi::execCLOSE_COMREQ(Signal* signal)
     }
   }
 
-  if (failNo != 0) 
+  if (requestType != CloseComReqConf::RT_NO_REPLY)
   {
+    ndbassert((requestType == CloseComReqConf::RT_API_FAILURE) ||
+              ((requestType == CloseComReqConf::RT_NODE_FAILURE) &&
+               (failNo != 0)));
     jam();
-    signal->theData[0] = userRef;
-    signal->theData[1] = failNo;
+    CloseComReqConf* closeComConf = (CloseComReqConf *)signal->getDataPtrSend();
+    closeComConf->xxxBlockRef = userRef;
+    closeComConf->requestType = requestType;
+    closeComConf->failNo = failNo;
+
+    /* Note assumption that noOfNodes and theNodes
+     * bitmap is not trampled above 
+     */
     sendSignal(QMGR_REF, GSN_CLOSE_COMCONF, signal, 19, JBA);
   }
 }
@@ -502,14 +634,6 @@ void Cmvmi::execOPEN_COMREQ(Signal* signal)
 	   && c_error_9000_nodes_mask.get(tStartingNode)))
 #endif
     {
-      if (globalData.theStartLevel != NodeState::SL_STARTED &&
-          (getNodeInfo(tStartingNode).m_type != NodeInfo::DB &&
-           getNodeInfo(tStartingNode).m_type != NodeInfo::MGM))
-      {
-        jam();
-        goto done;
-      }
-
       globalTransporterRegistry.do_connect(tStartingNode);
       globalTransporterRegistry.setIOState(tStartingNode, HaltIO);
       
@@ -534,7 +658,6 @@ void Cmvmi::execOPEN_COMREQ(Signal* signal)
 	    && c_error_9000_nodes_mask.get(i))
 	  continue;
 #endif
-	
 	globalTransporterRegistry.do_connect(i);
 	globalTransporterRegistry.setIOState(i, HaltIO);
 	
@@ -545,8 +668,8 @@ void Cmvmi::execOPEN_COMREQ(Signal* signal)
     }
   }
   
-done:  
-  if (userRef != 0) {
+  if (userRef != 0)
+  {
     jam(); 
     signal->theData[0] = tStartingNode;
     signal->theData[1] = tData2;
@@ -579,7 +702,6 @@ void Cmvmi::execDISCONNECT_REP(Signal *signal)
 {
   const DisconnectRep * const rep = (DisconnectRep *)&signal->theData[0];
   const Uint32 hostId = rep->nodeId;
-  const Uint32 errNo  = rep->err;
   
   jamEntry();
 
@@ -607,38 +729,17 @@ void Cmvmi::execCONNECT_REP(Signal *signal){
   globalData.m_nodeInfo[hostId].m_version = 0;
   globalData.m_nodeInfo[hostId].m_mysql_version = 0;
   
-  if(type == NodeInfo::DB || globalData.theStartLevel >= NodeState::SL_STARTED){
-    jam();
-    
-    /**
-     * Inform QMGR that client has connected
-     */
-
-    signal->theData[0] = hostId;
-    sendSignal(QMGR_REF, GSN_CONNECT_REP, signal, 1, JBA);
-  } else if(globalData.theStartLevel == NodeState::SL_CMVMI ||
-            globalData.theStartLevel == NodeState::SL_STARTING) {
-    jam();
-    /**
-     * Someone connected before start was finished
-     */
-    if(type == NodeInfo::MGM){
-      jam();
-      signal->theData[0] = hostId;
-      sendSignal(QMGR_REF, GSN_CONNECT_REP, signal, 1, JBA);
-    } else {
-      /**
-       * Dont allow api nodes to connect
-       */
-      ndbout_c("%d %d %d", hostId, type, globalData.theStartLevel);
-      abort();
-      globalTransporterRegistry.do_disconnect(hostId);
-    }
-  }
+  /**
+   * Inform QMGR that client has connected
+   */
   
+  signal->theData[0] = hostId;
+  sendSignal(QMGR_REF, GSN_CONNECT_REP, signal, 1, JBA);
+
   /* Automatically subscribe events for MGM nodes.
    */
-  if(type == NodeInfo::MGM){
+  if(type == NodeInfo::MGM)
+  {
     jam();
     globalTransporterRegistry.setIOState(hostId, NoHalt);
   }
@@ -787,7 +888,6 @@ void Cmvmi::execSTOP_ORD(Signal* signal)
 
 void
 Cmvmi::execSTART_ORD(Signal* signal) {
-
   StartOrd * const startOrd = (StartOrd *)&signal->theData[0];
   jamEntry();
   
@@ -827,26 +927,26 @@ Cmvmi::execSTART_ORD(Signal* signal) {
     return;
   }
   
-  if(globalData.theStartLevel == NodeState::SL_NOTHING){
+  if(globalData.theStartLevel == NodeState::SL_NOTHING)
+  {
     jam();
-    globalData.theStartLevel = NodeState::SL_CMVMI;
-    /**
-     * Open connections to management servers
-     */
-    for(unsigned int i = 1; i < MAX_NODES; i++ ){
-      if (getNodeInfo(i).m_type == NodeInfo::MGM){ 
-        if(!globalTransporterRegistry.is_connected(i)){
-          globalTransporterRegistry.do_connect(i);
-          globalTransporterRegistry.setIOState(i, NoHalt);
-        }
+
+    for(unsigned int i = 1; i < MAX_NODES; i++ )
+    {
+      if (getNodeInfo(i).m_type == NodeInfo::MGM)
+      {
+        jam();
+        globalTransporterRegistry.do_connect(i);
       }
     }
 
+    globalData.theStartLevel = NodeState::SL_CMVMI;
     EXECUTE_DIRECT(QMGR, GSN_START_ORD, signal, 1);
     return ;
   }
   
-  if(globalData.theStartLevel == NodeState::SL_CMVMI){
+  if(globalData.theStartLevel == NodeState::SL_CMVMI)
+  {
     jam();
 
     if(m_ctx.m_config.lockPagesInMainMemory() == 2)
@@ -854,12 +954,16 @@ Cmvmi::execSTART_ORD(Signal* signal) {
       int res = NdbMem_MemLockAll(1);
       if(res != 0)
       {
-	g_eventLogger.warning("Failed to memlock pages");
-	warningEvent("Failed to memlock pages");
+        char buf[100];
+        BaseString::snprintf(buf, sizeof(buf),
+                             "Failed to memlock pages, error: %d (%s)",
+                             errno, strerror(errno));
+        g_eventLogger->warning("%s", buf);
+        warningEvent("%s", buf);
       }
       else
       {
-	g_eventLogger.info("Locked future allocations");
+        g_eventLogger->info("Locked future allocations");
       }
     }
     
@@ -874,12 +978,16 @@ Cmvmi::execSTART_ORD(Signal* signal) {
     // Disconnect all nodes as part of the system restart. 
     // We need to ensure that we are starting up
     // without any connected nodes.   
-    for(unsigned int i = 1; i < MAX_NODES; i++ ){
-      if (i != getOwnNodeId() && getNodeInfo(i).m_type != NodeInfo::MGM){
+    for(unsigned int i = 1; i < MAX_NODES; i++ )
+    {
+      if (i != getOwnNodeId() && getNodeInfo(i).m_type != NodeInfo::MGM)
+      {
         globalTransporterRegistry.do_disconnect(i);
         globalTransporterRegistry.setIOState(i, HaltIO);
       }
     }
+
+    CRASH_INSERTION(9994);
     
     /**
      * Start running startphases
@@ -1027,15 +1135,15 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
   {
     SubscriberPtr ptr;
     subscribers.first(ptr);  
-    g_eventLogger.info("List subscriptions:");
+    g_eventLogger->info("List subscriptions:");
     while(ptr.i != RNIL)
     {
-      g_eventLogger.info("Subscription: %u, nodeId: %u, ref: 0x%x",
-                         ptr.i,  refToNode(ptr.p->blockRef), ptr.p->blockRef);
+      g_eventLogger->info("Subscription: %u, nodeId: %u, ref: 0x%x",
+                          ptr.i,  refToNode(ptr.p->blockRef), ptr.p->blockRef);
       for(Uint32 i = 0; i < LogLevel::LOGLEVEL_CATEGORIES; i++)
       {
         Uint32 level = ptr.p->logLevel.getLogLevel((LogLevel::EventCategory)i);
-        g_eventLogger.info("Category %u Level %u", i, level);
+        g_eventLogger->info("Category %u Level %u", i, level);
       }
       subscribers.next(ptr);
     }
@@ -1064,8 +1172,10 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
     else
     {
       if (rl.m_min || rl.m_curr || rl.m_max)
-	infoEvent("Resource %d min: %d max: %d curr: %d",
-		  id, rl.m_min, rl.m_max, rl.m_curr);
+      {
+        infoEvent("Resource %d min: %d max: %d curr: %d",
+                  id, rl.m_min, rl.m_max, rl.m_curr);
+      }
     }
 
     if (len == 3)
@@ -1077,7 +1187,31 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
     }
     return;
   }
-  
+  if (arg == DumpStateOrd::CmvmiSchedulerExecutionTimer)
+  {
+    Uint32 exec_time = signal->theData[1];
+    globalEmulatorData.theConfiguration->schedulerExecutionTimer(exec_time);
+  }
+  if (arg == DumpStateOrd::CmvmiSchedulerSpinTimer)
+  {
+    Uint32 spin_time = signal->theData[1];
+    globalEmulatorData.theConfiguration->schedulerSpinTimer(spin_time);
+  } 
+  if (arg == DumpStateOrd::CmvmiRealtimeScheduler)
+  {
+    bool realtime_on = signal->theData[1];
+    globalEmulatorData.theConfiguration->realtimeScheduler(realtime_on);
+  }
+  if (arg == DumpStateOrd::CmvmiExecuteLockCPU)
+  {
+    Uint32 exec_cpu_id = signal->theData[1];
+    globalEmulatorData.theConfiguration->executeLockCPU(exec_cpu_id);
+  }
+  if (arg == DumpStateOrd::CmvmiMaintLockCPU)
+  {
+    Uint32 maint_cpu_id = signal->theData[1];
+    globalEmulatorData.theConfiguration->maintLockCPU(maint_cpu_id);
+  }
   if (arg == DumpStateOrd::CmvmiSetRestartOnErrorInsert)
   {
     if(signal->getLength() == 1)
@@ -1126,6 +1260,16 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
     sendSignal(reference(), GSN_TESTSIG, signal, 8, JBB, ptr, 2);
   }
 
+  if (arg == DumpStateOrd::CmvmiTestLongSig)
+  {
+    /* Forward as GSN_TESTSIG to self */
+    Uint32 numArgs= signal->length() - 1;
+    memmove(signal->getDataPtrSend(), 
+            signal->getDataPtrSend() + 1, 
+            numArgs << 2);
+    sendSignal(reference(), GSN_TESTSIG, signal, numArgs, JBB);
+  }
+
 #ifdef ERROR_INSERT
   if (arg == 9000 || arg == 9002)
   {
@@ -1151,6 +1295,20 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
     }
     c_error_9000_nodes_mask.clear();
   }
+
+  if (arg == 9004 && signal->getLength() == 2)
+  {
+    SET_ERROR_INSERT_VALUE(9004);
+    c_error_9000_nodes_mask.clear();
+    c_error_9000_nodes_mask.set(signal->theData[1]);
+  }
+
+  if (arg == 9004 && signal->getLength() == 2)
+  {
+    SET_ERROR_INSERT_VALUE(9004);
+    c_error_9000_nodes_mask.clear();
+    c_error_9000_nodes_mask.set(signal->theData[1]);
+  }
 #endif
 
 #ifdef VM_TRACE
@@ -1225,6 +1383,376 @@ Cmvmi::execNODE_START_REP(Signal* signal)
 
 BLOCK_FUNCTIONS(Cmvmi)
 
+void
+Cmvmi::startFragmentedSend(Signal* signal,
+                           Uint32 variant,
+                           Uint32 numSigs,
+                           NodeReceiverGroup rg)
+{
+  Uint32* sigData = signal->getDataPtrSend();
+  const Uint32 sigLength = 6;
+  const Uint32 sectionWords = 240;
+  Uint32 sectionData[ sectionWords ];
+  
+  for (Uint32 i = 0; i < sectionWords; i++)
+    sectionData[ i ] = i;
+  
+  const Uint32 secCount = 1; 
+  LinearSectionPtr ptr[3];
+  ptr[0].sz = sectionWords;
+  ptr[0].p = &sectionData[0];
+
+  for (Uint32 i = 0; i < numSigs; i++)
+  {
+    sigData[0] = variant;
+    sigData[1] = 31;
+    sigData[2] = 0;
+    sigData[3] = 1; // print
+    sigData[4] = 0;
+    sigData[5] = sectionWords;
+    
+    if ((i & 1) == 0)
+    {
+      DEBUG("Starting linear fragmented send (" << i + 1
+            << "/" << numSigs << ")");
+
+      /* Linear send */
+      /* Todo : Avoid reading from invalid stackptr in CONTINUEB */
+      sendFragmentedSignal(rg,
+                           GSN_TESTSIG,
+                           signal,
+                           sigLength,
+                           JBB,
+                           ptr,
+                           secCount,
+                           TheEmptyCallback,
+                           90); // messageSize
+    }
+    else
+    {
+      /* Segmented send */
+      DEBUG("Starting segmented fragmented send (" << i + 1
+            << "/" << numSigs << ")");
+      Ptr<SectionSegment> segPtr;
+      ndbrequire(import(segPtr, sectionData, sectionWords));
+      SectionHandle handle(this, segPtr.i);
+      
+      sendFragmentedSignal(rg,
+                           GSN_TESTSIG,
+                           signal,
+                           sigLength,
+                           JBB,
+                           &handle,
+                           TheEmptyCallback,
+                           90); // messageSize
+    }
+  }
+}
+
+void
+Cmvmi::testNodeFailureCleanupCallback(Signal* signal, Uint32 data, Uint32 elementsCleaned)
+{
+  DEBUG("testNodeFailureCleanupCallback");
+  DEBUG("Data : " << data 
+        << " elementsCleaned : " << elementsCleaned);
+
+  debugPrintFragmentCounts();
+
+  Uint32 variant = data & 0xffff;
+  Uint32 testType = (data >> 16) & 0xffff;
+
+  DEBUG("Sending trigger(" << testType 
+        << ") variant " << variant 
+        << " to self to cleanup any fragments that arrived "
+        << "before send was cancelled");
+
+  Uint32* sigData = signal->getDataPtrSend();
+  sigData[0] = variant;
+  sigData[1] = testType;
+  sendSignal(reference(), GSN_TESTSIG, signal, 2, JBB);
+  
+  return; 
+}
+
+void 
+Cmvmi::testFragmentedCleanup(Signal* signal, SectionHandle* handle, Uint32 testType, Uint32 variant)
+{
+  DEBUG("TestType " << testType << " variant " << variant);
+  debugPrintFragmentCounts();
+
+  /* Variants : 
+   *     Local fragmented send   Multicast fragmented send
+   * 0 : Immediate cleanup       Immediate cleanup
+   * 1 : Continued cleanup       Immediate cleanup
+   * 2 : Immediate cleanup       Continued cleanup
+   * 3 : Continued cleanup       Continued cleanup
+   */
+  const Uint32 NUM_VARIANTS = 4;
+  if (variant >= NUM_VARIANTS)
+  {
+    DEBUG("Unsupported variant");
+    releaseSections(*handle);
+    return;
+  }
+
+  /* Test from ndb_mgm with
+   * <node(s)> DUMP 2605 0 30 
+   * 
+   * Use
+   * <node(s)> DUMP 2605 0 39 to get fragment resource usage counts
+   * Use
+   * <node(s)> DUMP 2601 to get segment usage counts in clusterlog
+   */
+  if (testType == 30)
+  {
+    /* Send the first fragment of a fragmented signal to self
+     * Receiver will allocate assembly hash entries
+     * which must be freed when node failure cleanup
+     * executes later
+     */
+    const Uint32 sectionWords = 240;
+    Uint32 sectionData[ sectionWords ];
+
+    for (Uint32 i = 0; i < sectionWords; i++)
+      sectionData[ i ] = i;
+
+    const Uint32 secCount = 1; 
+    LinearSectionPtr ptr[3];
+    ptr[0].sz = sectionWords;
+    ptr[0].p = &sectionData[0];
+
+    /* Send signal with testType == 31 */
+    NodeReceiverGroup me(reference());
+    Uint32* sigData = signal->getDataPtrSend();
+    const Uint32 sigLength = 6;
+    const Uint32 numPartialSigs = 4; 
+    /* Not too many as CMVMI's fragInfo hash is limited size */
+    // TODO : Consider making it debug-larger to get 
+    // more coverage on CONTINUEB path
+
+    for (Uint32 i = 0; i < numPartialSigs; i++)
+    {
+      /* Fill in messy TESTSIG format */
+      sigData[0] = variant;
+      sigData[1] = 31;
+      sigData[2] = 0;
+      sigData[3] = 0; // print
+      sigData[4] = 0;
+      sigData[5] = sectionWords;
+      
+      FragmentSendInfo fsi;
+      
+      DEBUG("Sending first fragment to self");
+      sendFirstFragment(fsi,
+                        me,
+                        GSN_TESTSIG,
+                        signal,
+                        sigLength,
+                        JBB,
+                        ptr,
+                        secCount,
+                        90); // FragmentLength
+
+      DEBUG("Cancelling remainder to free internal section");
+      fsi.m_status = FragmentSendInfo::SendCancelled;
+      sendNextLinearFragment(signal, fsi);
+    };
+
+    /* Ok, now send short signal with testType == 32
+     * to trigger 'remote-side' actions in middle of
+     * multiple fragment assembly
+     */
+    sigData[0] = variant;
+    sigData[1] = 32;
+
+    DEBUG("Sending node fail trigger to self");
+    sendSignal(me, GSN_TESTSIG, signal, 2, JBB);
+    return;
+  }
+
+  if (testType == 31)
+  {
+    /* Just release sections - execTESTSIG() has shown sections received */
+    releaseSections(*handle);
+    return;
+  }
+
+  if (testType == 32)
+  {
+    /* 'Remote side' trigger to clean up fragmented signal resources */
+    BlockReference senderRef = signal->getSendersBlockRef();
+    Uint32 sendingNode = refToNode(senderRef);
+    
+    /* Start sending some linear and fragmented responses to the
+     * sender, to exercise frag-send cleanup code when we execute
+     * node-failure later
+     */
+    DEBUG("Starting fragmented send using continueB back to self");
+
+    NodeReceiverGroup sender(senderRef);
+    startFragmentedSend(signal, variant, 6, sender);
+
+    debugPrintFragmentCounts();
+
+    Uint32 cbData= (((Uint32) 33) << 16) | variant;
+    Callback cb = { safe_cast(&Cmvmi::testNodeFailureCleanupCallback),
+                    cbData };
+
+    Callback* cbPtr = NULL;
+
+    bool passCallback = variant & 1;
+
+    if (passCallback)
+    {
+      DEBUG("Running simBlock failure code WITH CALLBACK for node " 
+            << sendingNode);
+      cbPtr = &cb;
+    }
+    else
+    {
+      DEBUG("Running simBlock failure code IMMEDIATELY (no callback) for node "
+            << sendingNode);
+      cbPtr = &TheEmptyCallback;
+    }
+
+    Uint32 elementsCleaned = simBlockNodeFailure(signal, sendingNode, *cbPtr);
+    
+    DEBUG("Elements cleaned by call : " << elementsCleaned);
+
+    debugPrintFragmentCounts();
+
+    if (! passCallback)
+    {
+      DEBUG("Variant " << variant << " manually executing callback");
+      /* We call the callback inline here to continue processing */
+      testNodeFailureCleanupCallback(signal, 
+                                     cbData,
+                                     elementsCleaned);
+    }
+
+    return;
+  }
+
+  if (testType == 33)
+  {
+    /* Original side - receive cleanup trigger from 'remote' side
+     * after node failure cleanup performed there.  We may have
+     * fragments it managed to send before the cleanup completed
+     * so we'll get rid of them.
+     * This would not be necessary in reality as this node would
+     * be failed
+     */
+    Uint32 sendingNode = refToNode(signal->getSendersBlockRef());
+    DEBUG("Running simBlock failure code for node " << sendingNode);
+
+    Uint32 elementsCleaned = simBlockNodeFailure(signal, sendingNode);
+
+    DEBUG("Elements cleaned : " << elementsCleaned);
+
+    /* Should have no fragment resources in use now */
+    ndbrequire(debugPrintFragmentCounts() == 0);
+
+    /* Now use ReceiverGroup to multicast a fragmented signal to
+     * all database nodes
+     */
+    DEBUG("Starting to send fragmented continueB to all nodes inc. self : ");
+    NodeReceiverGroup allNodes(CMVMI, c_dbNodes);
+    
+    unsigned nodeId = 0;
+    while((nodeId = c_dbNodes.find(nodeId+1)) != BitmaskImpl::NotFound)
+    {
+      DEBUG("Node " << nodeId);
+    }
+
+    startFragmentedSend(signal, variant, 8, allNodes);
+
+    debugPrintFragmentCounts();
+
+    Uint32 cbData= (((Uint32) 34) << 16) | variant;
+    Callback cb = { safe_cast(&Cmvmi::testNodeFailureCleanupCallback),
+                    cbData };
+    
+    Callback* cbPtr = NULL;
+    
+    bool passCallback = variant & 2;
+
+    if (passCallback)
+    {
+      DEBUG("Running simBlock failure code for self WITH CALLBACK (" 
+            << getOwnNodeId() << ")");
+      cbPtr= &cb;
+    }
+    else
+    {
+      DEBUG("Running simBlock failure code for self IMMEDIATELY (no callback) ("
+            << getOwnNodeId() << ")");
+      cbPtr= &TheEmptyCallback;
+    }
+    
+
+    /* Fragmented signals being sent will have this node removed
+     * from their receiver group, but will keep sending to the 
+     * other node(s).
+     * Other node(s) should therefore receive the complete signals.
+     * We will then receive only the first fragment of each of 
+     * the signals which must be removed later.
+     */
+    elementsCleaned = simBlockNodeFailure(signal, getOwnNodeId(), *cbPtr);
+
+    DEBUG("Elements cleaned : " << elementsCleaned);
+    
+    debugPrintFragmentCounts();
+
+    /* Callback will send a signal to self to clean up fragments that 
+     * were sent to self before the send was cancelled.  
+     * (Again, unnecessary in a 'real' situation
+     */
+    if (!passCallback)
+    {
+      DEBUG("Variant " << variant << " manually executing callback");
+
+      testNodeFailureCleanupCallback(signal,
+                                     cbData,
+                                     elementsCleaned);
+    }
+
+    return;
+  }
+  
+  if (testType == 34)
+  {
+    /* Cleanup fragments which were sent before send was cancelled. */
+    Uint32 elementsCleaned = simBlockNodeFailure(signal, getOwnNodeId());
+    
+    DEBUG("Elements cleaned " << elementsCleaned);
+    
+    /* All FragInfo should be clear, may still be sending some
+     * to other node(s)
+     */
+    debugPrintFragmentCounts();
+
+    DEBUG("Variant " << variant << " completed.");
+    
+    if (++variant < NUM_VARIANTS)
+    {
+      DEBUG("Re-executing with variant " << variant);
+      Uint32* sigData = signal->getDataPtrSend();
+      sigData[0] = variant;
+      sigData[1] = 30;
+      sendSignal(reference(), GSN_TESTSIG, signal, 2, JBB);
+    }
+//    else
+//    {
+//      // Infinite loop to test for leaks
+//       DEBUG("Back to zero");
+//       Uint32* sigData = signal->getDataPtrSend();
+//       sigData[0] = 0;
+//       sigData[1] = 30;
+//       sendSignal(reference(), GSN_TESTSIG, signal, 2, JBB);
+//    }
+  }
+}
+
 static Uint32 g_print;
 static LinearSectionPtr g_test[3];
 
@@ -1247,6 +1775,8 @@ Cmvmi::execTESTSIG(Signal* signal){
   g_print = signal->theData[3];
 //  Uint32 returnCount = signal->theData[4];
   Uint32 * secSizes = &signal->theData[5];
+
+  SectionHandle handle(this, signal);
   
   if(g_print){
     SignalLoggerManager::printSignalHeader(stdout, 
@@ -1262,10 +1792,10 @@ Cmvmi::execTESTSIG(Signal* signal){
     }
     fprintf(stdout, "\n");
     
-    for(i = 0; i<signal->header.m_noOfSections; i++){
+    for(i = 0; i<handle.m_cnt; i++){
       SegmentedSectionPtr ptr(0,0,0);
       ndbout_c("-- Section %d --", i);
-      signal->getSection(ptr, i);
+      handle.getSection(ptr, i);
       ndbrequire(ptr.p != 0);
       print(ptr, stdout);
       ndbrequire(ptr.sz == secSizes[i]);
@@ -1275,9 +1805,9 @@ Cmvmi::execTESTSIG(Signal* signal){
   /**
    * Validate length:s
    */
-  for(i = 0; i<signal->header.m_noOfSections; i++){
+  for(i = 0; i<handle.m_cnt; i++){
     SegmentedSectionPtr ptr;
-    signal->getSection(ptr, i);
+    handle.getSection(ptr, i);
     ndbrequire(ptr.p != 0);
     ndbrequire(ptr.sz == secSizes[i]);
   }
@@ -1287,7 +1817,7 @@ Cmvmi::execTESTSIG(Signal* signal){
    */
   if (testType == 20) {
     if (signal->theData[4] == 0) {
-      releaseSections(signal);
+      releaseSections(handle);
       return;
     }
     signal->theData[4]--;
@@ -1297,6 +1827,16 @@ Cmvmi::execTESTSIG(Signal* signal){
   
   NodeReceiverGroup rg(CMVMI, c_dbNodes);
 
+  /**
+   * Testing SimulatedBlock fragment assembly cleanup
+   */
+  if ((testType >= 30) &&
+      (testType < 40))
+  {
+    testFragmentedCleanup(signal, &handle, testType, ref);
+    return;
+  }
+
   if(signal->getSendersBlockRef() == ref){
     /**
      * Signal from API (not via NodeReceiverGroup)
@@ -1311,31 +1851,34 @@ Cmvmi::execTESTSIG(Signal* signal){
   
   switch(testType){
   case 1:
-    sendSignal(ref, GSN_TESTSIG,  signal, signal->length(), JBB);      
+    sendSignal(ref, GSN_TESTSIG,  signal, signal->length(), JBB,
+	       &handle);
     break;
   case 2:
-    sendSignal(rg, GSN_TESTSIG,  signal, signal->length(), JBB);
+    sendSignal(rg, GSN_TESTSIG,  signal, signal->length(), JBB,
+	       &handle);
     break;
   case 3:
   case 4:{
     LinearSectionPtr ptr[3];
-    const Uint32 secs = signal->getNoOfSections();
+    const Uint32 secs = handle.m_cnt;
     for(i = 0; i<secs; i++){
       SegmentedSectionPtr sptr(0,0,0);
-      signal->getSection(sptr, i);
+      handle.getSection(sptr, i);
       ptr[i].sz = sptr.sz;
       ptr[i].p = new Uint32[sptr.sz];
       copy(ptr[i].p, sptr);
     }
     
     if(testType == 3){
-      sendSignal(ref, GSN_TESTSIG,  signal, signal->length(), JBB, ptr, secs); 
+      sendSignal(ref, GSN_TESTSIG, signal, signal->length(), JBB, ptr, secs);
     } else {
-      sendSignal(rg, GSN_TESTSIG,  signal, signal->length(), JBB, ptr, secs); 
+      sendSignal(rg, GSN_TESTSIG, signal, signal->length(), JBB, ptr, secs);
     }
     for(Uint32 i = 0; i<secs; i++){
       delete[] ptr[i].p;
     }
+    releaseSections(handle);
     break;
   }
   case 5:
@@ -1355,7 +1898,9 @@ Cmvmi::execTESTSIG(Signal* signal){
 		      signal,
 		      signal->length(),
 		      JBB,
+		      &handle,
 		      fragmentLength);
+
     int count = 1;
     while(fragSend.m_status != FragmentSendInfo::SendComplete){
       count++;
@@ -1368,10 +1913,10 @@ Cmvmi::execTESTSIG(Signal* signal){
   case 7:
   case 8:{
     LinearSectionPtr ptr[3];
-    const Uint32 secs = signal->getNoOfSections();
+    const Uint32 secs = handle.m_cnt;
     for(i = 0; i<secs; i++){
       SegmentedSectionPtr sptr(0,0,0);
-      signal->getSection(sptr, i);
+      handle.getSection(sptr, i);
       ptr[i].sz = sptr.sz;
       ptr[i].p = new Uint32[sptr.sz];
       copy(ptr[i].p, sptr);
@@ -1406,6 +1951,7 @@ Cmvmi::execTESTSIG(Signal* signal){
     for(i = 0; i<secs; i++){
       delete[] ptr[i].p;
     }
+    releaseSections(handle);
     break;
   }
   case 9:
@@ -1419,12 +1965,14 @@ Cmvmi::execTESTSIG(Signal* signal){
       m_callBack.m_callbackData = 9;
       sendFragmentedSignal(ref,
 			   GSN_TESTSIG, signal, signal->length(), JBB, 
+			   &handle,
 			   m_callBack,
 			   fragmentLength);
     } else {
       m_callBack.m_callbackData = 10;
       sendFragmentedSignal(rg,
 			   GSN_TESTSIG, signal, signal->length(), JBB, 
+			   &handle,
 			   m_callBack,
 			   fragmentLength);
     }
@@ -1433,11 +1981,11 @@ Cmvmi::execTESTSIG(Signal* signal){
   case 11:
   case 12:{
 
-    const Uint32 secs = signal->getNoOfSections();
+    const Uint32 secs = handle.m_cnt;
     memset(g_test, 0, sizeof(g_test));
     for(i = 0; i<secs; i++){
       SegmentedSectionPtr sptr(0,0,0);
-      signal->getSection(sptr, i);
+      handle.getSection(sptr, i);
       g_test[i].sz = sptr.sz;
       g_test[i].p = new Uint32[sptr.sz];
       copy(g_test[i].p, sptr);
