@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008-2010 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <my_pthread.h>
@@ -37,6 +40,7 @@
 #include <mgmapi_config_parameters.h>
 
 int global_flag_skip_invalidate_cache = 0;
+int global_flag_skip_waiting_for_clean_cache = 0;
 //#define DEBUG_REG
 
 // Just a C wrapper for threadMain
@@ -66,6 +70,7 @@ ClusterMgr::ClusterMgr(TransporterFacade & _facade):
   noOfConnectedNodes= 0;
   theClusterMgrThread= 0;
   m_connect_count = 0;
+  m_auto_reconnect = -1;
   m_cluster_state = CS_waiting_for_clean_cache;
   DBUG_VOID_RETURN;
 }
@@ -124,9 +129,9 @@ ClusterMgr::startThread() {
   
   theClusterMgrThread = NdbThread_Create(runClusterMgr_C,
                                          (void**)this,
-                                         32768,
+                                         0, // default stack size
                                          "ndb_clustermgr",
-                                         NDB_THREAD_PRIO_LOW);
+                                         NDB_THREAD_PRIO_HIGH);
   NdbMutex_Unlock(clusterMgrThreadMutex);
 }
 
@@ -134,7 +139,7 @@ void
 ClusterMgr::doStop( ){
   DBUG_ENTER("ClusterMgr::doStop");
   NdbMutex_Lock(clusterMgrThreadMutex);
-  if(theStop){
+  if(theStop == 1){
     NdbMutex_Unlock(clusterMgrThreadMutex);
     DBUG_VOID_RETURN;
   }
@@ -240,11 +245,14 @@ ClusterMgr::threadMain( ){
     if (m_cluster_state == CS_waiting_for_clean_cache &&
         theFacade.m_globalDictCache)
     {
-      theFacade.m_globalDictCache->lock();
-      unsigned sz= theFacade.m_globalDictCache->get_size();
-      theFacade.m_globalDictCache->unlock();
-      if (sz)
-        goto next;
+      if (!global_flag_skip_waiting_for_clean_cache)
+      {
+        theFacade.m_globalDictCache->lock();
+        unsigned sz= theFacade.m_globalDictCache->get_size();
+        theFacade.m_globalDictCache->unlock();
+        if (sz)
+          goto next;
+      }
       m_cluster_state = CS_waiting_for_first_connect;
     }
 
@@ -482,8 +490,11 @@ ClusterMgr::execNF_COMPLETEREP(const Uint32 * theData){
   const NodeId nodeId = nfComp->failedNodeId;
   assert(nodeId > 0 && nodeId < MAX_NDB_NODES);
   
-  theFacade.ReportNodeFailureComplete(nodeId);
-  theNodes[nodeId].nfCompleteRep = true;
+  if (theNodes[nodeId].nfCompleteRep == false)
+  {
+    theFacade.ReportNodeFailureComplete(nodeId);
+    theNodes[nodeId].nfCompleteRep = true;
+  }
 }
 
 void
@@ -553,8 +564,7 @@ ClusterMgr::reportNodeFailed(NodeId nodeId, bool disconnect){
     theFacade.ReportNodeDead(nodeId);
   }
   
-  theNode.nfCompleteRep = false;
-  if(noOfAliveNodes == 0)
+  if (noOfConnectedNodes == 0)
   {
     if (!global_flag_skip_invalidate_cache &&
         theFacade.m_globalDictCache)
@@ -565,6 +575,15 @@ ClusterMgr::reportNodeFailed(NodeId nodeId, bool disconnect){
       m_connect_count ++;
       m_cluster_state = CS_waiting_for_clean_cache;
     }
+
+    if (m_auto_reconnect == 0)
+    {
+      theStop = 2;
+    }
+  }
+  theNode.nfCompleteRep = false;
+  if(noOfAliveNodes == 0)
+  {
     NFCompleteRep rep;
     for(Uint32 i = 1; i<MAX_NDB_NODES; i++){
       if(theNodes[i].defined && theNodes[i].nfCompleteRep == false){
@@ -593,7 +612,7 @@ ArbitMgr::ArbitMgr(TransporterFacade & _fac)
 
   theInputTimeout = 0;
   theInputFull = false;
-  memset(&theInputFull, 0, sizeof(theInputFull));
+  memset(&theInputBuffer, 0, sizeof(theInputBuffer));
   theState = StateInit;
 
   memset(&theStartReq, 0, sizeof(theStartReq));
@@ -634,7 +653,9 @@ ArbitMgr::doStart(const Uint32* theData)
   aSignal.init(GSN_ARBIT_STARTREQ, theData);
   sendSignalToThread(aSignal);
   theThread = NdbThread_Create(
-    runArbitMgr_C, (void**)this, 32768, "ndb_arbitmgr",
+    runArbitMgr_C, (void**)this,
+    0, // default stack size
+    "ndb_arbitmgr",
     NDB_THREAD_PRIO_HIGH);
   NdbMutex_Unlock(theThreadMutex);
 }
