@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008-2010 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef NdbEventOperationImpl_H
 #define NdbEventOperationImpl_H
@@ -372,10 +375,10 @@ public:
   int readBlobParts(char* buf, NdbBlob* blob,
                     Uint32 part, Uint32 count, Uint16* lenLoc);
   int receive_event();
-  const bool tableNameChanged() const;
-  const bool tableFrmChanged() const;
-  const bool tableFragmentationChanged() const;
-  const bool tableRangeListChanged() const;
+  bool tableNameChanged() const;
+  bool tableFrmChanged() const;
+  bool tableFragmentationChanged() const;
+  bool tableRangeListChanged() const;
   Uint64 getGCI();
   Uint32 getAnyValue() const;
   Uint64 getLatestGCI();
@@ -414,31 +417,10 @@ public:
   Uint32 m_oid;
 
   /*
-    m_node_bit_mask keeps track of which ndb nodes have reference to
-    an event op
-
-    - add    - TE_ACTIVE
-    - remove - TE_STOP, TE_NODE_FAILURE, TE_CLUSTER_FAILURE
-
-    TE_NODE_FAILURE and TE_CLUSTER_FAILURE are created as events
-    and added to all event ops listed as active or pending delete
-    in m_dropped_ev_op using insertDataL, includeing the blob
-    event ops referenced by a regular event op.
-    - NdbEventBuffer::report_node_failure
-    - NdbEventBuffer::completeClusterFailed
-
-    TE_ACTIVE is sent from the kernel on initial execute/start of the
-    event op, but is also internally generetad on node connect like
-    TE_NODE_FAILURE and TE_CLUSTER_FAILURE
-    - NdbEventBuffer::report_node_connected
-
-    when m_node_bit_mask becomes clear, the kernel reference is
-    removed from m_ref_count
-
-    node id 0 is used to denote that cluster has a reference
-   */
-
-  Bitmask<(unsigned int)_NDB_NODE_BITMASK_SIZE> m_node_bit_mask;
+    when parsed gci > m_stop_gci it is safe to drop operation
+    as kernel will not have any more references
+  */
+  Uint64 m_stop_gci;
 
   /*
     m_ref_count keeps track of outstanding references to an event
@@ -528,12 +510,12 @@ public:
   int insertDataL(NdbEventOperationImpl *op,
 		  const SubTableData * const sdata, Uint32 len,
 		  LinearSectionPtr ptr[3]);
-  void execSUB_GCP_COMPLETE_REP(const SubGcpCompleteRep * const, Uint32 len);
+  void execSUB_GCP_COMPLETE_REP(const SubGcpCompleteRep * const, Uint32 len,
+                                int complete_cluster_failure= 0);
   void complete_outof_order_gcis();
   
   void report_node_connected(Uint32 node_id);
-  void report_node_failure(Uint32 node_id);
-  void completeClusterFailed();
+  void report_node_failure_completed(Uint32 node_id);
 
   // used by user thread 
   Uint64 getLatestGCI();
@@ -547,7 +529,7 @@ public:
 
   NdbEventOperationImpl* getGCIEventOperations(Uint32* iter,
                                                Uint32* event_types);
-  void deleteUsedEventOperations();
+  void deleteUsedEventOperations(Uint64 last_consumed_gci);
 
   NdbEventOperationImpl *move_data();
 
@@ -586,8 +568,14 @@ public:
 #endif
 
   Ndb *m_ndb;
-  Uint64 m_latestGCI;           // latest "handover" GCI
+
+  // "latest gci" variables updated in receiver thread
+  Uint64 m_latestGCI;           // latest GCI completed in order
   Uint64 m_latest_complete_GCI; // latest complete GCI (in case of outof order)
+  Uint64 m_highest_sub_gcp_complete_GCI; // highest gci seen in api
+  // "latest gci" variables updated in user thread
+  Uint64 m_latest_poll_GCI; // latest gci handed over to user thread
+
   bool m_startup_hack;
 
   NdbMutex *m_mutex;
@@ -664,6 +652,8 @@ private:
   void complete_bucket(Gci_container*);
   bool find_max_known_gci(Uint64 * res) const;
   void resize_known_gci();
+
+  Bitmask<(unsigned int)_NDB_NODE_BITMASK_SIZE> m_alive_node_bit_mask;
 };
 
 inline
