@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <my_pthread.h>
@@ -60,6 +62,21 @@ static int indexToNumber(int index)
  * Call back functions
  *****************************************************************************/
 
+void* ndb_thread_add_thread_id(void *param)
+{
+  return NULL;
+}
+
+void *ndb_thread_remove_thread_id(void *param)
+{
+  return NULL;
+}
+
+void ndb_thread_fill_thread_object(void *param, uint *len, my_bool server)
+{
+  *len = 0;
+}
+
 void
 reportError(void * callbackObj, NodeId nodeId,
 	    TransporterError errorCode, const char *info)
@@ -352,7 +369,7 @@ execute(void * callbackObj, SignalHeader * const header,
        {
 	 Uint32* send= tSignal.getDataPtrSend();
 	 memcpy(send, theData, tSignal.getLength() << 2);
-	 ((SubGcpCompleteAck*)send)->rep.senderRef = 
+	 CAST_PTR(SubGcpCompleteAck, send)->rep.senderRef = 
 	   numberToRef(API_CLUSTERMGR, theFacade->theOwnId);
 	 Uint32 ref= header->theSendersBlockRef;
 	 Uint32 aNodeId= refToNode(ref);
@@ -362,6 +379,16 @@ execute(void * callbackObj, SignalHeader * const header,
        }
        break;
      }
+     case GSN_TAKE_OVERTCCONF:
+     {
+       /**
+	* Report
+	*/
+       NdbApiSignal tSignal(* header);
+       tSignal.setDataPtr(theData);
+       theFacade->for_each(&tSignal, ptr);
+       return;
+     }
      default:
        break;
        
@@ -736,6 +763,23 @@ TransporterFacade::init(Uint32 nodeId, const ndb_mgm_configuration* props)
   if (!iter.get(CFG_BATCH_SIZE, &batch_size)) {
     m_batch_size= batch_size;
   }
+
+  Uint32 auto_reconnect=1;
+  iter.get(CFG_AUTO_RECONNECT, &auto_reconnect);
+
+  const char * priospec = 0;
+  if (iter.get(CFG_HB_THREAD_PRIO, &priospec) == 0)
+  {
+    NdbThread_SetHighPrioProperties(priospec);
+  }
+
+  /**
+   * Keep value it set before connect (overriding config)
+   */
+  if (theClusterMgr->m_auto_reconnect == -1)
+  {
+    theClusterMgr->m_auto_reconnect = auto_reconnect;
+  }
   
   Uint32 timeout = 120000;
   iter.first();
@@ -757,13 +801,13 @@ TransporterFacade::init(Uint32 nodeId, const ndb_mgm_configuration* props)
 
   theReceiveThread = NdbThread_Create(runReceiveResponse_C,
                                       (void**)this,
-                                      32768,
+                                      0, // default stack size
                                       "ndb_receive",
                                       NDB_THREAD_PRIO_LOW);
 
   theSendThread = NdbThread_Create(runSendRequest_C,
                                    (void**)this,
-                                   32768,
+                                   0, // default stack size
                                    "ndb_send",
                                    NDB_THREAD_PRIO_LOW);
   theClusterMgr->startThread();
@@ -801,7 +845,7 @@ TransporterFacade::connected()
     if (m_threads.getInUse(i)){
       void * obj = m_threads.m_objectExecute[i].m_object;
       NodeStatusFunction RegPC = m_threads.m_statusFunction[i];
-      (*RegPC) (obj, numberToRef(indexToNumber(i), theOwnId), true, true);
+      (*RegPC) (obj, numberToRef(indexToNumber(i), theOwnId), NS_CONNECTED);
     }
   }
   DBUG_VOID_RETURN;
@@ -825,7 +869,7 @@ TransporterFacade::ReportNodeDead(NodeId tNodeId)
     if (m_threads.getInUse(i)){
       void * obj = m_threads.m_objectExecute[i].m_object;
       NodeStatusFunction RegPC = m_threads.m_statusFunction[i];
-      (*RegPC) (obj, tNodeId, false, false);
+      (*RegPC) (obj, tNodeId, NS_NODE_FAILED);
     }
   }
   DBUG_VOID_RETURN;
@@ -850,7 +894,7 @@ TransporterFacade::ReportNodeFailureComplete(NodeId tNodeId)
     if (m_threads.getInUse(i)){
       void * obj = m_threads.m_objectExecute[i].m_object;
       NodeStatusFunction RegPC = m_threads.m_statusFunction[i];
-      (*RegPC) (obj, tNodeId, false, true);
+      (*RegPC) (obj, tNodeId, NS_NODE_NF_COMPLETE);
     }
   }
   DBUG_VOID_RETURN;
@@ -873,7 +917,7 @@ TransporterFacade::ReportNodeAlive(NodeId tNodeId)
     if (m_threads.getInUse(i)){
       void * obj = m_threads.m_objectExecute[i].m_object;
       NodeStatusFunction RegPC = m_threads.m_statusFunction[i];
-      (*RegPC) (obj, tNodeId, true, false);
+      (*RegPC) (obj, tNodeId, NS_NODE_ALIVE);
     }
   }
 }
@@ -906,7 +950,7 @@ TransporterFacade::open(void* objRef,
     DBUG_RETURN(r);
 #if 1
   if (theOwnId > 0) {
-    (*statusFun)(objRef, numberToRef(r, theOwnId), true, true);
+    (*statusFun)(objRef, numberToRef(r, theOwnId), NS_CONNECTED);
   }
 #endif
   DBUG_RETURN(r);
@@ -1032,15 +1076,18 @@ TransporterFacade::sendSignal(NdbApiSignal * aSignal, NodeId aNode){
 }
 
 int
-TransporterFacade::sendSignalUnCond(NdbApiSignal * aSignal, NodeId aNode){
+TransporterFacade::sendSignalUnCond(NdbApiSignal * aSignal, 
+                                    NodeId aNode,
+                                    Uint32 prio){
   Uint32* tDataPtr = aSignal->getDataPtrSend();
+  assert(prio <= 1);
 #ifdef API_TRACE
   if(setSignalLog() && TRACE_GSN(aSignal->theVerId_signalNumber)){
     Uint32 tmp = aSignal->theSendersBlockRef;
     aSignal->theSendersBlockRef = numberToRef(tmp, theOwnId);
     LinearSectionPtr ptr[3];
     signalLogger.sendSignal(* aSignal,
-			    0,
+			    prio,
 			    tDataPtr,
 			    aNode, ptr, 0);
     signalLogger.flushSignalLog();
@@ -1051,7 +1098,7 @@ TransporterFacade::sendSignalUnCond(NdbApiSignal * aSignal, NodeId aNode){
          (aSignal->theLength <= 25) &&
          (aSignal->theReceiversBlockNumber != 0));
   SendStatus ss = theTransporterRegistry->prepareSend(aSignal, 
-						      0, 
+						      prio, 
 						      tDataPtr,
 						      aNode, 
 						      0);
@@ -1544,11 +1591,32 @@ SignalSender::sendSignal(Uint16 nodeId, const SimpleSignal * s){
     signalLogger.flushSignalLog();
   }
 #endif
-  assert(getNodeInfo(nodeId).m_api_reg_conf == true ||
-         s->readSignalNumber() == GSN_API_REGREQ);
-  return theFacade->theTransporterRegistry->prepareSend(&s->header,
-							1, // JBB
-							&s->theData[0],
-							nodeId, 
-							&s->ptr[0]);
+  
+  SendStatus ss = 
+    theFacade->theTransporterRegistry->prepareSend(&s->header,
+                                                   1, // JBB
+                                                   &s->theData[0],
+                                                   nodeId, 
+                                                   &s->ptr[0]);
+
+  if (ss == SEND_OK)
+  {
+    assert(getNodeInfo(nodeId).m_api_reg_conf == true ||
+           s->readSignalNumber() == GSN_API_REGREQ);
+    theFacade->forceSend(m_blockNo);
+  }
+
+  return ss;
+}
+
+void
+TransporterFacade::set_auto_reconnect(int val)
+{
+  theClusterMgr->m_auto_reconnect = val;
+}
+
+int
+TransporterFacade::get_auto_reconnect() const
+{
+  return theClusterMgr->m_auto_reconnect;
 }
