@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2004, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <my_pthread.h>
@@ -31,14 +33,14 @@
 #include <mgmapi_internal.h>
 #include <md5_hash.hpp>
 
-#include <EventLogger.hpp>
-EventLogger g_eventLogger;
-
 #include <NdbMutex.h>
 #ifdef VM_TRACE
 NdbMutex *ndb_print_state_mutex= NULL;
 #endif
 
+#include <EventLogger.hpp>
+extern EventLogger *g_eventLogger;
+
 static int g_ndb_connection_count = 0;
 
 /*
@@ -109,7 +111,9 @@ int Ndb_cluster_connection::start_connect_thread(int (*connect_callback)(void))
     DBUG_PRINT("info",("starting thread"));
     m_impl.m_connect_thread= 
       NdbThread_Create(run_ndb_cluster_connection_connect_thread,
-		       (void**)&m_impl, 32768, "ndb_cluster_connection",
+		       (void**)&m_impl,
+                       0, // default stack size
+                       "ndb_cluster_connection",
 		       NDB_THREAD_PRIO_LOW);
   }
   else if (r < 0)
@@ -181,6 +185,28 @@ Ndb_cluster_connection_impl::get_next_node(Ndb_cluster_connection_node_iter &ite
   return node.id;
 }
 
+Uint32
+Ndb_cluster_connection_impl::get_next_alive_node(Ndb_cluster_connection_node_iter &iter)
+{
+  Uint32 id;
+
+  TransporterFacade *tp = m_impl.m_transporter_facade;
+  if (tp == 0 || tp->ownId() == 0)
+    return 0;
+
+  while ((id = get_next_node(iter)))
+  {
+    tp->lock_mutex();
+    if (tp->get_node_alive(id) != 0)
+    {
+      tp->unlock_mutex();
+      return id;
+    }
+    tp->unlock_mutex();
+  }
+  return 0;
+}
+
 unsigned
 Ndb_cluster_connection::no_db_nodes()
 {
@@ -265,6 +291,16 @@ unsigned Ndb_cluster_connection::get_connect_count() const
   return m_impl.get_connect_count();
 }
 
+int Ndb_cluster_connection::get_latest_error() const
+{
+  return m_impl.m_latest_error;
+}
+
+const char *Ndb_cluster_connection::get_latest_error_msg() const
+{
+  return m_impl.m_latest_error_msg.c_str();
+}
+
 /*
  * Ndb_cluster_connection_impl
  */
@@ -280,18 +316,33 @@ Ndb_cluster_connection_impl(const char *
     m_run_connect_thread(0),
     m_event_add_drop_mutex(0),
     m_latest_trans_gci(0),
-    m_first_ndb_object(0)
+    m_first_ndb_object(0),
+    m_latest_error_msg(),
+    m_latest_error(0)
 {
   DBUG_ENTER("Ndb_cluster_connection");
   DBUG_PRINT("enter",("Ndb_cluster_connection this=0x%lx", (long) this));
 
+  NdbMutex_Lock(g_ndb_connection_mutex);
+  if(g_ndb_connection_count++ == 0)
+  {
+    NdbColumnImpl::create_pseudo_columns();
+    g_eventLogger->createConsoleHandler();
+    g_eventLogger->setCategory("NdbApi");
+    g_eventLogger->enable(Logger::LL_ON, Logger::LL_ERROR);
+    /*
+      Disable repeated message handling as it interfers
+      with mysqld logging, in which case messages come out
+      of order.  Same applies for regular ndbapi user.
+    */
+    g_eventLogger->setRepeatFrequency(0);
+  }
+  NdbMutex_Unlock(g_ndb_connection_mutex);
+
   if (!m_event_add_drop_mutex)
     m_event_add_drop_mutex= NdbMutex_Create();
   m_new_delete_ndb_mutex = NdbMutex_Create();
-  
-  g_eventLogger.createConsoleHandler();
-  g_eventLogger.setCategory("NdbApi");
-  g_eventLogger.enable(Logger::LL_ON, Logger::LL_ERROR);
+
 
   m_connect_thread= 0;
   m_connect_callback= 0;
@@ -304,8 +355,11 @@ Ndb_cluster_connection_impl(const char *
     new ConfigRetriever(connect_string, NDB_VERSION, NODE_TYPE_API);
   if (m_config_retriever->hasError())
   {
-    printf("Could not initialize handle to management server: %s\n",
-	   m_config_retriever->getErrorString());
+    m_latest_error= 1;
+    m_latest_error_msg.assfmt
+      ("Could not initialize handle to management server: %s",
+       m_config_retriever->getErrorString());
+    printf("%s\n", get_latest_error_msg());
     delete m_config_retriever;
     m_config_retriever= 0;
   }
@@ -327,36 +381,8 @@ Ndb_cluster_connection_impl(const char *
       new TransporterFacade(m_main_connection->m_impl.m_globalDictCache);
   }
 
-  NdbMutex_Lock(g_ndb_connection_mutex);
-  if(g_ndb_connection_count++ == 0){
-    NdbDictionary::Column::FRAGMENT= 
-      NdbColumnImpl::create_pseudo("NDB$FRAGMENT");
-    NdbDictionary::Column::FRAGMENT_FIXED_MEMORY= 
-      NdbColumnImpl::create_pseudo("NDB$FRAGMENT_FIXED_MEMORY");
-    NdbDictionary::Column::FRAGMENT_VARSIZED_MEMORY= 
-      NdbColumnImpl::create_pseudo("NDB$FRAGMENT_VARSIZED_MEMORY");
-    NdbDictionary::Column::ROW_COUNT= 
-      NdbColumnImpl::create_pseudo("NDB$ROW_COUNT");
-    NdbDictionary::Column::COMMIT_COUNT= 
-      NdbColumnImpl::create_pseudo("NDB$COMMIT_COUNT");
-    NdbDictionary::Column::ROW_SIZE=
-      NdbColumnImpl::create_pseudo("NDB$ROW_SIZE");
-    NdbDictionary::Column::RANGE_NO= 
-      NdbColumnImpl::create_pseudo("NDB$RANGE_NO");
-    NdbDictionary::Column::DISK_REF= 
-      NdbColumnImpl::create_pseudo("NDB$DISK_REF");
-    NdbDictionary::Column::RECORDS_IN_RANGE= 
-      NdbColumnImpl::create_pseudo("NDB$RECORDS_IN_RANGE");
-    NdbDictionary::Column::ROWID= 
-      NdbColumnImpl::create_pseudo("NDB$ROWID");
-    NdbDictionary::Column::ROW_GCI= 
-      NdbColumnImpl::create_pseudo("NDB$ROW_GCI");
-    NdbDictionary::Column::ANY_VALUE= 
-      NdbColumnImpl::create_pseudo("NDB$ANY_VALUE");
-    NdbDictionary::Column::COPY_ROWID= 
-      NdbColumnImpl::create_pseudo("NDB$COPY_ROWID");
-  }
-  NdbMutex_Unlock(g_ndb_connection_mutex);
+
+
 
   DBUG_VOID_RETURN;
 }
@@ -364,6 +390,22 @@ Ndb_cluster_connection_impl(const char *
 Ndb_cluster_connection_impl::~Ndb_cluster_connection_impl()
 {
   DBUG_ENTER("~Ndb_cluster_connection");
+
+  if (m_first_ndb_object != 0)
+  {
+    g_eventLogger->warning("Deleting Ndb_cluster_connection with Ndb-object"
+                           " not deleted");
+    Ndb * p = m_first_ndb_object;
+    printf("this: %p Ndb-object(s): ", (Ndb_cluster_connection*)this);
+    while (p)
+    {
+      printf("%p ", p);
+      p = p->theImpl->m_next_ndb_object;
+    }
+    printf("\n");
+    fflush(stdout);
+  }
+
   if (m_transporter_facade != 0)
   {
     m_transporter_facade->stop_instance();
@@ -401,34 +443,9 @@ Ndb_cluster_connection_impl::~Ndb_cluster_connection_impl()
     free(m_name);
 
   NdbMutex_Lock(g_ndb_connection_mutex);
-  if(--g_ndb_connection_count == 0){
-    delete NdbDictionary::Column::FRAGMENT; 
-    delete NdbDictionary::Column::FRAGMENT_FIXED_MEMORY;
-    delete NdbDictionary::Column::FRAGMENT_VARSIZED_MEMORY;
-    delete NdbDictionary::Column::ROW_COUNT;
-    delete NdbDictionary::Column::COMMIT_COUNT;
-    delete NdbDictionary::Column::ROW_SIZE;
-    delete NdbDictionary::Column::RANGE_NO;
-    delete NdbDictionary::Column::DISK_REF;
-    delete NdbDictionary::Column::RECORDS_IN_RANGE;
-    delete NdbDictionary::Column::ROWID;
-    delete NdbDictionary::Column::ROW_GCI;
-    delete NdbDictionary::Column::ANY_VALUE;
-    NdbDictionary::Column::FRAGMENT= 0;
-    NdbDictionary::Column::FRAGMENT_FIXED_MEMORY= 0;
-    NdbDictionary::Column::FRAGMENT_VARSIZED_MEMORY= 0;
-    NdbDictionary::Column::ROW_COUNT= 0;
-    NdbDictionary::Column::COMMIT_COUNT= 0;
-    NdbDictionary::Column::ROW_SIZE= 0;
-    NdbDictionary::Column::RANGE_NO= 0;
-    NdbDictionary::Column::DISK_REF= 0;
-    NdbDictionary::Column::RECORDS_IN_RANGE= 0;
-    NdbDictionary::Column::ROWID= 0;
-    NdbDictionary::Column::ROW_GCI= 0;
-    NdbDictionary::Column::ANY_VALUE= 0;
-
-    delete NdbDictionary::Column::COPY_ROWID;
-    NdbDictionary::Column::COPY_ROWID = 0;
+  if(--g_ndb_connection_count == 0)
+  {
+    NdbColumnImpl::destory_pseudo_columns();
   }
   NdbMutex_Unlock(g_ndb_connection_mutex);
 
@@ -682,11 +699,27 @@ int Ndb_cluster_connection::connect(int no_retries, int retry_delay_in_seconds,
   DBUG_ENTER("Ndb_cluster_connection::connect");
   do {
     if (m_impl.m_config_retriever == 0)
+    {
+      if (!m_impl.m_latest_error)
+      {
+        m_impl.m_latest_error= 1;
+        m_impl.m_latest_error_msg.assign
+          ("Ndb_cluster_connection init error: m_impl.m_config_retriever==0");
+      }
+      DBUG_PRINT("exit", ("no m_impl.m_config_retriever, ret: -1"));
       DBUG_RETURN(-1);
+    }
     if (m_impl.m_config_retriever->do_connect(no_retries,
 					      retry_delay_in_seconds,
 					      verbose))
+    {
+      char buf[1024];
+      m_impl.m_latest_error= 1;
+      m_impl.m_latest_error_msg.assfmt
+        ("Connect using '%s' timed out", get_connectstring(buf, sizeof(buf)));
+      DBUG_PRINT("exit", ("mgmt server not up yet, ret: 1"));
       DBUG_RETURN(1); // mgmt server not up yet
+    }
 
     Uint32 nodeId = m_impl.m_config_retriever->allocNodeId(4/*retries*/,
 							   3/*delay*/);
@@ -700,6 +733,7 @@ int Ndb_cluster_connection::connect(int no_retries, int retry_delay_in_seconds,
     if (m_impl.init_nodes_vector(nodeId, *props))
     {
       ndbout_c("Ndb_cluster_connection::connect: malloc failure");
+      DBUG_PRINT("exit", ("malloc failure, ret: -1"));
       DBUG_RETURN(-1);
     }
 
@@ -719,15 +753,20 @@ int Ndb_cluster_connection::connect(int no_retries, int retry_delay_in_seconds,
 
     ndb_mgm_destroy_configuration(props);
     m_impl.m_transporter_facade->connected();
+    m_impl.m_latest_error= 0;
+    m_impl.m_latest_error_msg.assign("");
+    DBUG_PRINT("exit", ("connect ok, ret: 0"));
     DBUG_RETURN(0);
   } while(0);
   
-  ndbout << "Configuration error: ";
   const char* erString = m_impl.m_config_retriever->getErrorString();
   if (erString == 0) {
     erString = "No error specified!";
   }
-  ndbout << erString << endl;
+  m_impl.m_latest_error= 1;
+  m_impl.m_latest_error_msg.assfmt("Configuration error: %s", erString);
+  ndbout << get_latest_error_msg() << endl;
+  DBUG_PRINT("exit", ("connect failed, '%s' ret: -1", erString));
   DBUG_RETURN(-1);
 }
 
@@ -771,6 +810,12 @@ Ndb_cluster_connection::get_next_node(Ndb_cluster_connection_node_iter &iter)
   return m_impl.get_next_node(iter);
 }
 
+unsigned int 
+Ndb_cluster_connection::get_next_alive_node(Ndb_cluster_connection_node_iter &iter)
+{
+  return m_impl.get_next_alive_node(iter);
+}
+
 unsigned
 Ndb_cluster_connection::get_active_ndb_objects() const
 {
@@ -783,5 +828,17 @@ int Ndb_cluster_connection::set_timeout(int timeout_ms)
                              timeout_ms);
 }
 
+int
+Ndb_cluster_connection::get_auto_reconnect() const
+{
+  return m_impl.m_transporter_facade->get_auto_reconnect();
+}
+
+void
+Ndb_cluster_connection::set_auto_reconnect(int value)
+{
+  m_impl.m_transporter_facade->set_auto_reconnect(value);
+}
+
 template class Vector<Ndb_cluster_connection_impl::Node>;
 
