@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "Restore.hpp"
 #include <NdbTCP.h>
@@ -28,51 +30,216 @@
 
 #include "../../../../sql/ha_ndbcluster_tables.h"
 extern NdbRecordPrintFormat g_ndbrecord_print_format;
+extern bool ga_skip_unknown_objects;
+extern bool ga_skip_broken_objects;
 
 Uint16 Twiddle16(Uint16 in); // Byte shift 16-bit data
 Uint32 Twiddle32(Uint32 in); // Byte shift 32-bit data
 Uint64 Twiddle64(Uint64 in); // Byte shift 64-bit data
 
-bool
-BackupFile::Twiddle(const AttributeDesc* attr_desc, AttributeData* attr_data, Uint32 arraySize){
-  Uint32 i;
 
-  if(m_hostByteOrder)
-    return true;
-  
-  if(arraySize == 0){
-    arraySize = attr_desc->arraySize;
+/*
+  TwiddleUtil
+
+  Utility class used when swapping byteorder
+  of one attribute in a table
+
+*/
+
+class TwiddleUtil {
+  Uint32 m_twiddle_size;
+  Uint32 m_twiddle_array_size;
+public:
+  TwiddleUtil(); // Not implemented
+  TwiddleUtil(const TwiddleUtil&); // Not implemented
+
+  TwiddleUtil(const AttributeDesc * const attr_desc) {
+    const NdbDictionary::Column::Type attribute_type =
+      attr_desc->m_column->getType();
+
+    switch(attribute_type){
+    case NdbDictionary::Column::Datetime:
+      // Datetime is stored as 8x8, should be twiddled as 64 bit
+      assert(attr_desc->size == 8);
+      assert(attr_desc->arraySize == 8);
+      m_twiddle_size = 64;
+      m_twiddle_array_size = 1;
+      break;
+
+    case NdbDictionary::Column::Timestamp:
+      // Timestamp is stored as 4x8, should be twiddled as 32 bit
+      assert(attr_desc->size == 8);
+      assert(attr_desc->arraySize == 4);
+      m_twiddle_size = 32;
+      m_twiddle_array_size = 1;
+      break;
+
+    case NdbDictionary::Column::Blob:
+    case NdbDictionary::Column::Text:
+      if (attr_desc->m_column->getArrayType() ==
+          NdbDictionary::Column::ArrayTypeFixed)
+      {
+        // Length of fixed size blob which is stored in first 64 bit's
+        // has to be twiddled, the remaining byte stream left as is
+        assert(attr_desc->size == 8);
+        assert(attr_desc->arraySize > 8);
+        m_twiddle_size = 64;
+        m_twiddle_array_size = 1;
+        break;
+      }
+      // Fallthrough for blob/text with ArrayTypeVar
+
+    default:
+      // Default twiddling parameters
+      m_twiddle_size = attr_desc->size;
+      m_twiddle_array_size = attr_desc->arraySize;
+      break;
+    }
+
+    assert(m_twiddle_array_size);
+    assert(m_twiddle_size);
   }
-  
-  switch(attr_desc->size){
-  case 8:
-    
-    return true;
-  case 16:
-    for(i = 0; i<arraySize; i++){
-      attr_data->u_int16_value[i] = Twiddle16(attr_data->u_int16_value[i]);
-    }
-    return true;
-  case 32:
-    for(i = 0; i<arraySize; i++){
-      attr_data->u_int32_value[i] = Twiddle32(attr_data->u_int32_value[i]);
-    }
-    return true;
-  case 64:
-    for(i = 0; i<arraySize; i++){
-      // allow unaligned
-      char* p = (char*)&attr_data->u_int64_value[i];
-      Uint64 x;
-      memcpy(&x, p, sizeof(Uint64));
-      x = Twiddle64(x);
-      memcpy(p, &x, sizeof(Uint64));
-    }
-    return true;
-  default:
-    return false;
-  } // switch
 
-} // Twiddle
+  bool is_aligned (void* data_ptr) const {
+    switch (m_twiddle_size){
+    case 8:
+      // Always aligned
+      return true;
+      break;
+    case 16:
+      return ((((size_t)data_ptr) & 1) == 0);
+      break;
+    case 32:
+      return ((((size_t)data_ptr) & 3) == 0);
+      break;
+    case 64:
+      return ((((size_t)data_ptr) & 7) == 0);
+      break;
+    default:
+      abort();
+    }
+  }
+
+  void twiddle_aligned(void* const data_ptr) const {
+    // Make sure the data pointer is properly aligned
+    assert(is_aligned(data_ptr));
+
+    switch(m_twiddle_size){
+    case 8:
+      // Nothing to swap
+      break;
+    case 16:
+    {
+      Uint16* ptr = (Uint16*)data_ptr;
+      for (Uint32 i = 0; i < m_twiddle_array_size; i++){
+        *ptr = Twiddle16(*ptr);
+        ptr++;
+      }
+      break;
+    }
+    case 32:
+    {
+      Uint32* ptr = (Uint32*)data_ptr;
+      for (Uint32 i = 0; i < m_twiddle_array_size; i++){
+        *ptr = Twiddle32(*ptr);
+        ptr++;
+      }
+      break;
+    }
+    case 64:
+    {
+      Uint64* ptr = (Uint64*)data_ptr;
+      for (Uint32 i = 0; i < m_twiddle_array_size; i++){
+        *ptr = Twiddle64(*ptr);
+        ptr++;
+      }
+      break;
+    }
+    default:
+      abort();
+    } // switch
+  }
+};
+
+
+/*
+  BackupFile::twiddle_attribute
+
+  Swap the byte order of one attribute whose data may or may not
+  be properly aligned for the current datatype
+
+*/
+
+void
+BackupFile::twiddle_atribute(const AttributeDesc * const attr_desc,
+                             AttributeData* attr_data)
+{
+  TwiddleUtil map(attr_desc);
+
+  // Check if data is aligned properly
+  void* data_ptr = (char*)attr_data->void_value;
+  Uint32 data_sz = attr_desc->getSizeInBytes();
+  bool aligned= map.is_aligned(data_ptr);
+  if (!aligned)
+  {
+    // The pointer is not properly aligned, copy the data
+    // to aligned memory before twiddling
+    m_twiddle_buffer.assign(data_ptr, data_sz);
+    data_ptr = m_twiddle_buffer.get_data();
+  }
+
+  // Swap the byteorder of the aligned data
+  map.twiddle_aligned(data_ptr);
+
+  if (!aligned)
+  {
+    // Copy data back from aligned memory
+    memcpy(attr_data->void_value,
+           m_twiddle_buffer.get_data(),
+           data_sz);
+  }
+}
+
+
+/*
+  BackupFile::Twiddle
+
+  Swap the byteorder for one attribute if it was stored
+  in different byteorder than current host
+
+*/
+
+bool
+BackupFile::Twiddle(const AttributeDesc * const attr_desc,
+                    AttributeData* attr_data)
+{
+  // Check parameters are not NULL
+  assert(attr_desc);
+  assert(attr_data);
+
+  // Make sure there is data to fiddle with
+  assert(!attr_data->null);
+  assert(attr_data->void_value);
+
+  if(unlikely(!m_hostByteOrder))
+  {
+    // The data file is not in host byte order, the
+    // attribute need byte order swapped
+    twiddle_atribute(attr_desc, attr_data);
+  }
+#ifdef VM_TRACE
+  else
+  {
+    // Increase test converage in debug mode by doing
+    // a double byte order swap to prove that both ways work
+    twiddle_atribute(attr_desc, attr_data);
+    twiddle_atribute(attr_desc, attr_data);
+  }
+#endif
+
+  return true;
+}
+
 
 FilteredNdbOut err(* new FileOutputStream(stderr), 0, 0);
 FilteredNdbOut info(* new FileOutputStream(stdout), 1, 1);
@@ -166,7 +333,8 @@ RestoreMetaData::readMetaTableDesc() {
   
   // Read section header 
   Uint32 sz = sizeof(sectionInfo) >> 2;
-  if (m_fileHeader.NdbVersion < NDBD_ROWID_VERSION)
+  if (m_fileHeader.NdbVersion < NDBD_ROWID_VERSION ||
+      isDrop6(m_fileHeader.NdbVersion))
   {
     sz = 2;
     sectionInfo[2] = htonl(DictTabInfo::UserTable);
@@ -251,8 +419,17 @@ RestoreMetaData::readMetaTableDesc() {
     break;
   }
   default:
-    err << "Unsupported table type!! " << sectionInfo[2] << endl;
-    return false;
+    if (ga_skip_unknown_objects)
+    {
+      info << "Skipping schema object with unknown table type "
+           << sectionInfo[2] << endl;
+      return true;
+    }
+    else
+    {
+      err << "Unsupported table type!! " << sectionInfo[2] << endl;
+      return false;
+    }
   }
   if (errcode)
   {
@@ -318,8 +495,12 @@ RestoreMetaData::markSysTables()
         strcmp(tableName, OLD_NDB_REP_DB "/def/" OLD_NDB_SCHEMA_TABLE) == 0 ||
         strcmp(tableName, NDB_REP_DB "/def/" NDB_APPLY_TABLE) == 0 ||
         strcmp(tableName, NDB_REP_DB "/def/" NDB_SCHEMA_TABLE)== 0 )
-
-      table->isSysTable = true;
+    {
+      table->m_isSysTable = true;
+      if (strcmp(tableName, "SYSTAB_0") == 0 ||
+          strcmp(tableName, "sys/def/SYSTAB_0") == 0)
+        table->m_isSYSTAB_0 = true;
+    }
   }
   for (i = 0; i < getNoOfTables(); i++) {
     TableS* blobTable = allTables[i];
@@ -334,8 +515,8 @@ RestoreMetaData::markSysTables()
       for (j = 0; j < getNoOfTables(); j++) {
         TableS* table = allTables[j];
         if (table->getTableId() == (Uint32) id1) {
-          if (table->isSysTable)
-            blobTable->isSysTable = true;
+          if (table->m_isSysTable)
+            blobTable->m_isSysTable = true;
           blobTable->m_main_table = table;
           blobTable->m_main_column_id = id2;
           break;
@@ -361,7 +542,6 @@ RestoreMetaData::fixBlobs()
     const Uint32 noOfBlobs = t.m_noOfBlobs;
     if (noOfBlobs == 0)
       continue;
-    const Uint32 noOfColumns = t.getNoOfColumns();
     Uint32 n = 0;
     Uint32 j;
     for (j = 0; n < noOfBlobs; j++) {
@@ -383,7 +563,24 @@ RestoreMetaData::fixBlobs()
           break;
         }
       }
-      assert(blobTable != NULL);
+      if (blobTable == NULL)
+      {
+        table->m_broken = true;
+        /* Corrupt backup, has main table, but no blob table */
+        err << "Table " << table->m_dictTable->getName()
+            << " has blob column " << j << " (" 
+            << c->m_name.c_str()
+            << ") with missing parts table in backup."
+            << endl;
+        if (ga_skip_broken_objects)
+        {
+          continue;
+        }
+        else
+        {
+          return false;
+        }
+      }
       assert(blobTable->m_dictTable != NULL);
       NdbTableImpl& bt = NdbTableImpl::getImpl(*blobTable->m_dictTable);
       const char* colName = c->m_blobVersion == 1 ? "DATA" : "NDB$DATA";
@@ -399,29 +596,26 @@ RestoreMetaData::fixBlobs()
 bool
 RestoreMetaData::readGCPEntry() {
 
-  Uint32 data[4];
+  BackupFormat::CtlFile::GCPEntry dst;
   
-  BackupFormat::CtlFile::GCPEntry * dst = 
-    (BackupFormat::CtlFile::GCPEntry *)&data[0];
-  
-  if(buffer_read(dst, 4, 4) != 4){
+  if(buffer_read(&dst, 1, sizeof(dst)) != sizeof(dst)){
     err << "readGCPEntry read error" << endl;
     return false;
   }
   
-  dst->SectionType = ntohl(dst->SectionType);
-  dst->SectionLength = ntohl(dst->SectionLength);
+  dst.SectionType = ntohl(dst.SectionType);
+  dst.SectionLength = ntohl(dst.SectionLength);
   
-  if(dst->SectionType != BackupFormat::GCP_ENTRY){
+  if(dst.SectionType != BackupFormat::GCP_ENTRY){
     err << "readGCPEntry invalid format" << endl;
     return false;
   }
   
-  dst->StartGCP = ntohl(dst->StartGCP);
-  dst->StopGCP = ntohl(dst->StopGCP);
+  dst.StartGCP = ntohl(dst.StartGCP);
+  dst.StopGCP = ntohl(dst.StopGCP);
   
-  m_startGCP = dst->StartGCP;
-  m_stopGCP = dst->StopGCP;
+  m_startGCP = dst.StartGCP;
+  m_stopGCP = dst.StopGCP;
   return true;
 }
 
@@ -476,11 +670,13 @@ TableS::TableS(Uint32 version, NdbTableImpl* tableImpl)
 {
   m_dictTable = tableImpl;
   m_noOfNullable = m_nullBitmaskSize = 0;
-  m_auto_val_id= ~(Uint32)0;
+  m_auto_val_attrib = 0;
   m_max_auto_val= 0;
   m_noOfRecords= 0;
   backupVersion = version;
-  isSysTable = false;
+  m_isSysTable = false;
+  m_isSYSTAB_0 = false;
+  m_broken = false;
   m_main_table = NULL;
   m_main_column_id = ~(Uint32)0;
   
@@ -491,7 +687,11 @@ TableS::TableS(Uint32 version, NdbTableImpl* tableImpl)
 TableS::~TableS()
 {
   for (Uint32 i= 0; i < allAttributesDesc.size(); i++)
+  {
+    if (allAttributesDesc[i]->parameter)
+      free(allAttributesDesc[i]->parameter);
     delete allAttributesDesc[i];
+  }
 }
 
 
@@ -500,8 +700,10 @@ bool
 RestoreMetaData::parseTableDescriptor(const Uint32 * data, Uint32 len)
 {
   NdbTableImpl* tableImpl = 0;
-  int ret = NdbDictInterface::parseTableInfo(&tableImpl, data, len, false,
-                                             m_fileHeader.NdbVersion);
+  int ret = NdbDictInterface::parseTableInfo
+    (&tableImpl, data, len, false,
+     isDrop6(m_fileHeader.NdbVersion) ? MAKE_VERSION(5,1,2) :
+     m_fileHeader.NdbVersion);
   
   if (ret != 0) {
     err << "parseTableInfo " << " failed" << endl;
@@ -534,6 +736,71 @@ RestoreDataIterator::RestoreDataIterator(const RestoreMetaData & md, void (* _fr
 {
   debug << "RestoreDataIterator constructor" << endl;
   setDataFile(md, 0);
+
+  m_bitfield_storage_len = 8192;
+  m_bitfield_storage_ptr = (Uint32*)malloc(4*m_bitfield_storage_len);
+  m_bitfield_storage_curr_ptr = m_bitfield_storage_ptr;
+  m_row_bitfield_len = 0;
+}
+
+RestoreDataIterator::~RestoreDataIterator()
+{
+  free_bitfield_storage();
+}
+
+void
+RestoreDataIterator::init_bitfield_storage(const NdbDictionary::Table* tab)
+{
+  Uint32 len = 0;
+  for (Uint32 i = 0; i<(Uint32)tab->getNoOfColumns(); i++)
+  {
+    if (tab->getColumn(i)->getType() == NdbDictionary::Column::Bit)
+    {
+      len += (tab->getColumn(i)->getLength() + 31) >> 5;
+    }
+  }
+
+  m_row_bitfield_len = len;
+}
+
+void
+RestoreDataIterator::reset_bitfield_storage()
+{
+  m_bitfield_storage_curr_ptr = m_bitfield_storage_ptr;
+}
+
+void
+RestoreDataIterator::free_bitfield_storage()
+{
+  if (m_bitfield_storage_ptr)
+    free(m_bitfield_storage_ptr);
+  m_bitfield_storage_ptr = 0;
+  m_bitfield_storage_curr_ptr = 0;
+  m_bitfield_storage_len = 0;
+}
+
+Uint32
+RestoreDataIterator::get_free_bitfield_storage() const
+{
+  return (m_bitfield_storage_ptr + m_bitfield_storage_len) - 
+    m_bitfield_storage_curr_ptr;
+}
+
+Uint32*
+RestoreDataIterator::get_bitfield_storage(Uint32 len)
+{
+  Uint32 * currptr = m_bitfield_storage_curr_ptr;
+  Uint32 * nextptr = currptr + len;
+  Uint32 * endptr = m_bitfield_storage_ptr + m_bitfield_storage_len;
+
+  if (nextptr <= endptr)
+  {
+    m_bitfield_storage_curr_ptr = nextptr;
+    return currptr;
+  }
+  
+  abort();
+  return 0;
 }
 
 TupleS & TupleS::operator=(const TupleS& tuple)
@@ -555,7 +822,7 @@ TableS * TupleS::getTable() const {
   return m_currentTable;
 }
 
-const AttributeDesc * TupleS::getDesc(int i) const {
+AttributeDesc * TupleS::getDesc(int i) const {
   return m_currentTable->allAttributesDesc[i];
 }
 
@@ -584,9 +851,315 @@ TupleS::prepareRecord(TableS & tab){
   return true;
 }
 
+static
+inline
+Uint8*
+pad(Uint8* src, Uint32 align, Uint32 bitPos)
+{
+  UintPtr ptr = UintPtr(src);
+  switch(align){
+  case DictTabInfo::aBit:
+  case DictTabInfo::a32Bit:
+  case DictTabInfo::a64Bit:
+  case DictTabInfo::a128Bit:
+    return (Uint8*)(((ptr + 3) & ~(UintPtr)3) + 4 * ((bitPos + 31) >> 5));
+charpad:
+  case DictTabInfo::an8Bit:
+  case DictTabInfo::a16Bit:
+    return src + 4 * ((bitPos + 31) >> 5);
+  default:
+#ifdef VM_TRACE
+    abort();
+#endif
+    goto charpad;
+  }
+}
+
+const TupleS *
+RestoreDataIterator::getNextTuple(int  & res)
+{
+  if (m_currentTable->backupVersion >= NDBD_RAW_LCP)
+  {
+    if (m_row_bitfield_len >= get_free_bitfield_storage())
+    {
+      /**
+       * Informing buffer reader that it does not need to cache
+       * "old" data here would be clever...
+       * But I can't find a good/easy way to do this
+       */
+      if (free_data_callback)
+        (*free_data_callback)();
+      reset_bitfield_storage();
+    }
+  }
+  
+  Uint32  dataLength = 0;
+  // Read record length
+  if (buffer_read(&dataLength, sizeof(dataLength), 1) != 1){
+    err << "getNextTuple:Error reading length  of data part" << endl;
+    res = -1;
+    return NULL;
+  } // if
+  
+  // Convert length from network byte order
+  dataLength = ntohl(dataLength);
+  const Uint32 dataLenBytes = 4 * dataLength;
+  
+  if (dataLength == 0) {
+    // Zero length for last tuple
+    // End of this data fragment
+    debug << "End of fragment" << endl;
+    res = 0;
+    return NULL;
+  } // if
+
+  // Read tuple data
+  void *_buf_ptr;
+  if (buffer_get_ptr(&_buf_ptr, 1, dataLenBytes) != dataLenBytes) {
+    err << "getNextTuple:Read error: " << endl;
+    res = -1;
+    return NULL;
+  }
+
+  Uint32 *buf_ptr = (Uint32*)_buf_ptr;
+  if (m_currentTable->backupVersion >= NDBD_RAW_LCP)
+  {
+    res = readTupleData_packed(buf_ptr, dataLength);
+  }
+  else
+  {
+    res = readTupleData_old(buf_ptr, dataLength);
+  }
+  
+  if (res)
+  {
+    return NULL;
+  }
+
+  m_count ++;  
+  res = 0;
+  return &m_tuple;
+} // RestoreDataIterator::getNextTuple
+
+TableS *
+RestoreDataIterator::getCurrentTable()
+{
+  return m_currentTable;
+}
+
 int
-RestoreDataIterator::readTupleData(Uint32 *buf_ptr, Uint32 *ptr,
-                                   Uint32 dataLength)
+RestoreDataIterator::readTupleData_packed(Uint32 *buf_ptr, 
+                                          Uint32 dataLength)
+{
+  Uint32 * ptr = buf_ptr;
+  /**
+   * Unpack READ_PACKED header
+   */
+  Uint32 rp = * ptr;
+  if(unlikely(!m_hostByteOrder))
+    rp = Twiddle32(rp);
+
+  AttributeHeader ah(rp);
+  assert(ah.getAttributeId() == AttributeHeader::READ_PACKED);
+  Uint32 bmlen = ah.getByteSize();
+  assert((bmlen & 3) == 0);
+  Uint32 bmlen32 = bmlen / 4;
+
+  /**
+   * Twiddle READ_BACKED header
+   */
+  if (!m_hostByteOrder)
+  {
+    for (Uint32 i = 0; i < 1 + bmlen32; i++)
+    {
+      ptr[i] = Twiddle32(ptr[i]);
+    }
+  }
+  
+  const NdbDictionary::Table* tab = m_currentTable->m_dictTable;
+  
+  // All columns should be present...
+  assert(((tab->getNoOfColumns() + 31) >> 5) <= (int)bmlen32);
+  
+  /**
+   * Iterate through attributes...
+   */
+  const Uint32 * bmptr = ptr + 1;
+  Uint8* src = (Uint8*)(bmptr + bmlen32);
+  Uint32 bmpos = 0;
+  Uint32 bitPos = 0;
+  for (Uint32 i = 0; i < (Uint32)tab->getNoOfColumns(); i++, bmpos++)
+  {
+    // All columns should be present
+    assert(BitmaskImpl::get(bmlen32, bmptr, bmpos));
+    const NdbColumnImpl & col = NdbColumnImpl::getImpl(* tab->getColumn(i));
+    AttributeData * attr_data = m_tuple.getData(i);
+    const AttributeDesc * attr_desc = m_tuple.getDesc(i);
+    if (col.getNullable())
+    {
+      bmpos++;
+      if (BitmaskImpl::get(bmlen32, bmptr, bmpos))
+      {
+        attr_data->null = true;
+        attr_data->void_value = NULL;
+        continue;
+      }
+    }
+    
+    attr_data->null = false;
+    
+    /**
+     * Handle padding
+     */
+    Uint32 align = col.m_orgAttrSize;
+    Uint32 attrSize = col.m_attrSize;
+    Uint32 array = col.m_arraySize;
+    Uint32 len = col.m_length;
+    Uint32 sz = attrSize * array;
+    Uint32 arrayType = col.m_arrayType;
+    
+    switch(align){
+    case DictTabInfo::aBit:{ // Bit
+      src = pad(src, 0, 0);
+      Uint32* src32 = (Uint32*)src;
+      
+      Uint32 len32 = (len + 31) >> 5;
+      Uint32* tmp = get_bitfield_storage(len32);
+      attr_data->null = false;
+      attr_data->void_value = tmp;
+      attr_data->size = 4*len32;
+      
+      if (m_hostByteOrder)
+      {
+        BitmaskImpl::getField(1 + len32, src32, bitPos, len, tmp);
+      }
+      else
+      {
+        Uint32 ii;
+        for (ii = 0; ii< (1 + len32); ii++)
+          src32[ii] = Twiddle32(src32[ii]);
+        BitmaskImpl::getField(1 + len32, (Uint32*)src, bitPos, len, tmp);
+        for (ii = 0; ii< (1 + len32); ii++)
+          src32[ii] = Twiddle32(src32[ii]);
+      }
+      
+      src += 4 * ((bitPos + len) >> 5);
+      bitPos = (bitPos + len) & 31;
+      goto next;
+    }
+    default:
+      src = pad(src, align, bitPos);
+    }
+    switch(arrayType){
+    case NDB_ARRAYTYPE_FIXED:
+      break;
+    case NDB_ARRAYTYPE_SHORT_VAR:
+      sz = 1 + src[0];
+      break;
+    case NDB_ARRAYTYPE_MEDIUM_VAR:
+      sz = 2 + src[0] + 256 * src[1];
+      break;
+    default:
+      abort();
+    }
+    
+    attr_data->void_value = src;
+    attr_data->size = sz;
+    
+    if(!Twiddle(attr_desc, attr_data))
+    {
+      return -1;
+    }
+    
+    /**
+     * Next
+     */
+    bitPos = 0;
+    src += sz;
+next:
+    (void)1;
+  }
+  return 0;
+}
+
+int
+RestoreDataIterator::readTupleData_old(Uint32 *buf_ptr, 
+                                       Uint32 dataLength)
+{
+  Uint32 * ptr = buf_ptr;
+  ptr += m_currentTable->m_nullBitmaskSize;
+  Uint32 i;
+  for(i= 0; i < m_currentTable->m_fixedKeys.size(); i++){
+    assert(ptr < buf_ptr + dataLength);
+ 
+    const Uint32 attrId = m_currentTable->m_fixedKeys[i]->attrId;
+
+    AttributeData * attr_data = m_tuple.getData(attrId);
+    const AttributeDesc * attr_desc = m_tuple.getDesc(attrId);
+
+    const Uint32 sz = attr_desc->getSizeInWords();
+
+    attr_data->null = false;
+    attr_data->void_value = ptr;
+    attr_data->size = 4*sz;
+
+    if(!Twiddle(attr_desc, attr_data))
+    {
+      return -1;
+    }
+    ptr += sz;
+  }
+  
+  for(i = 0; i < m_currentTable->m_fixedAttribs.size(); i++){
+    assert(ptr < buf_ptr + dataLength);
+
+    const Uint32 attrId = m_currentTable->m_fixedAttribs[i]->attrId;
+
+    AttributeData * attr_data = m_tuple.getData(attrId);
+    const AttributeDesc * attr_desc = m_tuple.getDesc(attrId);
+
+    const Uint32 sz = attr_desc->getSizeInWords();
+
+    attr_data->null = false;
+    attr_data->void_value = ptr;
+    attr_data->size = 4*sz;
+
+    if(!Twiddle(attr_desc, attr_data))
+    {
+      return -1;
+    }
+    
+    ptr += sz;
+  }
+
+  // init to NULL
+  for(i = 0; i < m_currentTable->m_variableAttribs.size(); i++){
+    const Uint32 attrId = m_currentTable->m_variableAttribs[i]->attrId;
+
+    AttributeData * attr_data = m_tuple.getData(attrId);
+    
+    attr_data->null = true;
+    attr_data->void_value = NULL;
+  }
+
+  int res;
+  if (!isDrop6(m_currentTable->backupVersion))
+  {
+    if ((res = readVarData(buf_ptr, ptr, dataLength)))
+      return res;
+  }
+  else
+  {
+    if ((res = readVarData_drop6(buf_ptr, ptr, dataLength)))
+      return res;
+  }
+
+  return 0;
+}
+
+int
+RestoreDataIterator::readVarData(Uint32 *buf_ptr, Uint32 *ptr,
+                                  Uint32 dataLength)
 {
   while (ptr + 2 < buf_ptr + dataLength)
   {
@@ -621,39 +1194,8 @@ RestoreDataIterator::readTupleData(Uint32 *buf_ptr, Uint32 *ptr,
     attr_data->void_value = &data->Data[0];
     attr_data->size = sz;
 
-    //if (m_currentTable->getTableId() >= 2) { ndbout << "var off=" << ptr-buf_ptr << " attrId=" << attrId << endl; }
-
-    /**
-     * Compute array size
-     */
-    const Uint32 arraySize = sz / (attr_desc->size / 8);
-    assert(arraySize <= attr_desc->arraySize);
-
     //convert the length of blob(v1) and text(v1)
-    if(!m_hostByteOrder
-        && (attr_desc->m_column->getType() == NdbDictionary::Column::Blob
-           || attr_desc->m_column->getType() == NdbDictionary::Column::Text)
-        && attr_desc->m_column->getArrayType() == NdbDictionary::Column::ArrayTypeFixed)
-    {
-      char* p = (char*)&attr_data->u_int64_value[0];
-      Uint64 x;
-      memcpy(&x, p, sizeof(Uint64));
-      x = Twiddle64(x);
-      memcpy(p, &x, sizeof(Uint64));
-    }
-
-    //convert datetime type
-    if(!m_hostByteOrder
-        && attr_desc->m_column->getType() == NdbDictionary::Column::Datetime)
-    {
-      char* p = (char*)&attr_data->u_int64_value[0];
-      Uint64 x;
-      memcpy(&x, p, sizeof(Uint64));
-      x = Twiddle64(x);
-      memcpy(p, &x, sizeof(Uint64));
-    }
-
-    if(!Twiddle(attr_desc, attr_data, attr_desc->arraySize))
+    if(!Twiddle(attr_desc, attr_data))
     {
       return -1;
     }
@@ -666,114 +1208,57 @@ RestoreDataIterator::readTupleData(Uint32 *buf_ptr, Uint32 *ptr,
   return 0;
 }
 
-const TupleS *
-RestoreDataIterator::getNextTuple(int  & res)
+
+int
+RestoreDataIterator::readVarData_drop6(Uint32 *buf_ptr, Uint32 *ptr,
+                                       Uint32 dataLength)
 {
-  Uint32  dataLength = 0;
-  // Read record length
-  if (buffer_read(&dataLength, sizeof(dataLength), 1) != 1){
-    err << "getNextTuple:Error reading length  of data part" << endl;
-    res = -1;
-    return NULL;
-  } // if
-  
-  // Convert length from network byte order
-  dataLength = ntohl(dataLength);
-  const Uint32 dataLenBytes = 4 * dataLength;
-  
-  if (dataLength == 0) {
-    // Zero length for last tuple
-    // End of this data fragment
-    debug << "End of fragment" << endl;
-    res = 0;
-    return NULL;
-  } // if
-
-  // Read tuple data
-  void *_buf_ptr;
-  if (buffer_get_ptr(&_buf_ptr, 1, dataLenBytes) != dataLenBytes) {
-    err << "getNextTuple:Read error: " << endl;
-    res = -1;
-    return NULL;
-  }
- 
-  //if (m_currentTable->getTableId() >= 2) { for (uint ii=0; ii<dataLenBytes; ii+=4) ndbout << "*" << hex << *(Uint32*)( (char*)_buf_ptr+ii ); ndbout << endl; }
-
-  Uint32 *buf_ptr = (Uint32*)_buf_ptr, *ptr = buf_ptr;
-  ptr += m_currentTable->m_nullBitmaskSize;
   Uint32 i;
-  for(i= 0; i < m_currentTable->m_fixedKeys.size(); i++){
-    assert(ptr < buf_ptr + dataLength);
- 
-    const Uint32 attrId = m_currentTable->m_fixedKeys[i]->attrId;
-
-    AttributeData * attr_data = m_tuple.getData(attrId);
-    const AttributeDesc * attr_desc = m_tuple.getDesc(attrId);
-
-    const Uint32 sz = attr_desc->getSizeInWords();
-
-    attr_data->null = false;
-    attr_data->void_value = ptr;
-    attr_data->size = 4*sz;
-
-    if(!Twiddle(attr_desc, attr_data))
-      {
-	res = -1;
-	return NULL;
-      }
-    ptr += sz;
-  }
-
-  for(i = 0; i < m_currentTable->m_fixedAttribs.size(); i++){
-    assert(ptr < buf_ptr + dataLength);
-
-    const Uint32 attrId = m_currentTable->m_fixedAttribs[i]->attrId;
-
-    AttributeData * attr_data = m_tuple.getData(attrId);
-    const AttributeDesc * attr_desc = m_tuple.getDesc(attrId);
-
-    const Uint32 sz = attr_desc->getSizeInWords();
-
-    attr_data->null = false;
-    attr_data->void_value = ptr;
-    attr_data->size = 4*sz;
-
-    //if (m_currentTable->getTableId() >= 2) { ndbout << "fix i=" << i << " off=" << ptr-buf_ptr << " attrId=" << attrId << endl; }
-    if(!m_hostByteOrder
-        && attr_desc->m_column->getType() == NdbDictionary::Column::Timestamp)
-      attr_data->u_int32_value[0] = Twiddle32(attr_data->u_int32_value[0]);
-
-    if(!Twiddle(attr_desc, attr_data))
-      {
-	res = -1;
-	return NULL;
-      }
-
-    ptr += sz;
-  }
-
-  // init to NULL
-  for(i = 0; i < m_currentTable->m_variableAttribs.size(); i++){
+  for (i = 0; i < m_currentTable->m_variableAttribs.size(); i++)
+  {
     const Uint32 attrId = m_currentTable->m_variableAttribs[i]->attrId;
 
     AttributeData * attr_data = m_tuple.getData(attrId);
+    const AttributeDesc * attr_desc = m_tuple.getDesc(attrId);
 
-    attr_data->null = true;
-    attr_data->void_value = NULL;
+    if(attr_desc->m_column->getNullable())
+    {
+      const Uint32 ind = attr_desc->m_nullBitIndex;
+      if(BitmaskImpl::get(m_currentTable->m_nullBitmaskSize, 
+                          buf_ptr,ind))
+      {
+        attr_data->null = true;
+        attr_data->void_value = NULL;
+        continue;
+      }
+    }
+
+    assert(ptr < buf_ptr + dataLength);
+
+    typedef BackupFormat::DataFile::VariableData VarData;
+    VarData * data = (VarData *)ptr;
+    Uint32 sz = ntohl(data->Sz);
+    Uint32 id = ntohl(data->Id);
+    assert(id == attrId);
+
+    attr_data->null = false;
+    attr_data->void_value = &data->Data[0];
+
+    if (!Twiddle(attr_desc, attr_data))
+    {
+      return -1;
+    }
+    ptr += (sz + 2);
   }
-
-  if ((res = readTupleData(buf_ptr, ptr, dataLength)))
-    return NULL;
-
-  m_count ++;  
-  res = 0;
-  return &m_tuple;
-} // RestoreDataIterator::getNextTuple
+  assert(ptr == buf_ptr + dataLength);
+  return 0;
+}
 
 BackupFile::BackupFile(void (* _free_data_callback)()) 
   : free_data_callback(_free_data_callback)
 {
-  m_file = 0;
+  memset(&m_file,0,sizeof(m_file));
+  m_file.file = -1;
   m_path[0] = 0;
   m_fileName[0] = 0;
 
@@ -787,29 +1272,32 @@ BackupFile::BackupFile(void (* _free_data_callback)())
 }
 
 BackupFile::~BackupFile(){
-  if(m_file != 0)
-    fclose(m_file);
+  if(m_file.file > 1)
+  {
+    azclose(&m_file);
+    memset(&m_file,0,sizeof(m_file));
+  }
   if(m_buffer != 0)
     free(m_buffer);
 }
 
 bool
 BackupFile::openFile(){
-  if(m_file != NULL){
-    fclose(m_file);
-    m_file = 0;
+  if(m_file.file > 1){
+    azclose(&m_file);
+    m_file.file = 0;
     m_file_size = 0;
     m_file_pos = 0;
   }
-  
+
   info.setLevel(254);
   info << "Opening file '" << m_fileName << "'\n";
-  m_file = fopen(m_fileName, "r");
+  int r= azopen(&m_file,m_fileName, O_RDONLY);
 
-  if (m_file)
+  if (r==0)
   {
     struct stat buf;
-    if (fstat(fileno(m_file), &buf) == 0)
+    if (fstat(m_file.file, &buf) == 0)
     {
       m_file_size = (Uint64)buf.st_size;
       info << "File size " << m_file_size << " bytes\n";
@@ -822,7 +1310,7 @@ BackupFile::openFile(){
     }
   }
 
-  return m_file != 0;
+  return r != 0;
 }
 
 Uint32 BackupFile::buffer_get_ptr_ahead(void **p_buf_ptr, Uint32 size, Uint32 nmemb)
@@ -833,9 +1321,13 @@ Uint32 BackupFile::buffer_get_ptr_ahead(void **p_buf_ptr, Uint32 size, Uint32 nm
     if (free_data_callback)
       (*free_data_callback)();
 
+    reset_buffers();
     memcpy(m_buffer, m_buffer_ptr, m_buffer_data_left);
 
-    size_t r = fread(((char *)m_buffer) + m_buffer_data_left, 1, m_buffer_sz - m_buffer_data_left, m_file);
+    int error;
+    size_t r = azread(&m_file,
+                      ((char *)m_buffer) + m_buffer_data_left,
+                      m_buffer_sz - m_buffer_data_left, &error);
     m_file_pos += r;
     m_buffer_data_left += r;
     m_buffer_ptr = m_buffer;
@@ -883,7 +1375,7 @@ BackupFile::setCtlFile(Uint32 nodeId, Uint32 backupId, const char * path){
   m_expectedFileHeader.FileType = BackupFormat::CTL_FILE;
 
   char name[PATH_MAX]; const Uint32 sz = sizeof(name);
-  BaseString::snprintf(name, sz, "BACKUP-%d.%d.ctl", backupId, nodeId);  
+  BaseString::snprintf(name, sz, "BACKUP-%u.%d.ctl", backupId, nodeId);  
   setName(path, name);
 }
 
@@ -894,7 +1386,7 @@ BackupFile::setDataFile(const BackupFile & bf, Uint32 no){
   m_expectedFileHeader.FileType = BackupFormat::DATA_FILE;
   
   char name[PATH_MAX]; const Uint32 sz = sizeof(name);
-  BaseString::snprintf(name, sz, "BACKUP-%d-%d.%d.Data", 
+  BaseString::snprintf(name, sz, "BACKUP-%u-%d.%d.Data", 
 	   m_expectedFileHeader.BackupId, no, m_nodeId);
   setName(bf.m_path, name);
 }
@@ -906,7 +1398,7 @@ BackupFile::setLogFile(const BackupFile & bf, Uint32 no){
   m_expectedFileHeader.FileType = BackupFormat::LOG_FILE;
   
   char name[PATH_MAX]; const Uint32 sz = sizeof(name);
-  BaseString::snprintf(name, sz, "BACKUP-%d.%d.log", 
+  BaseString::snprintf(name, sz, "BACKUP-%u.%d.log", 
 	   m_expectedFileHeader.BackupId, m_nodeId);
   setName(bf.m_path, name);
 }
@@ -934,13 +1426,21 @@ BackupFile::readHeader(){
     return false;
   }
   
-  if(buffer_read(&m_fileHeader, sizeof(m_fileHeader), 1) != 1){
+  Uint32 oldsz = sizeof(BackupFormat::FileHeader_pre_backup_version);
+  if(buffer_read(&m_fileHeader, oldsz, 1) != 1){
     err << "readDataFileHeader: Error reading header" << endl;
     return false;
   }
   
   // Convert from network to host byte order for platform compatibility
-  m_fileHeader.NdbVersion  = ntohl(m_fileHeader.NdbVersion);
+  /*
+    Due to some optimization going on when using gcc 4.2.3 we
+    have to read 'backup_version' into tmp variable. If
+    'm_fileHeader.BackupVersion' is used directly in the if statement
+    below it will have the wrong value.
+  */
+  Uint32 backup_version = ntohl(m_fileHeader.BackupVersion);
+  m_fileHeader.BackupVersion = backup_version;
   m_fileHeader.SectionType = ntohl(m_fileHeader.SectionType);
   m_fileHeader.SectionLength = ntohl(m_fileHeader.SectionLength);
   m_fileHeader.FileType = ntohl(m_fileHeader.FileType);
@@ -948,8 +1448,26 @@ BackupFile::readHeader(){
   m_fileHeader.BackupKey_0 = ntohl(m_fileHeader.BackupKey_0);
   m_fileHeader.BackupKey_1 = ntohl(m_fileHeader.BackupKey_1);
 
+  if (backup_version >= NDBD_RAW_LCP)
+  {
+    if (buffer_read(&m_fileHeader.NdbVersion, 
+                    sizeof(m_fileHeader) - oldsz, 1) != 1)
+    {
+      err << "readDataFileHeader: Error reading header" << endl;
+      return false;
+    }
+    
+    m_fileHeader.NdbVersion = ntohl(m_fileHeader.NdbVersion);
+    m_fileHeader.MySQLVersion = ntohl(m_fileHeader.MySQLVersion);
+  }
+  else
+  {
+    m_fileHeader.NdbVersion = m_fileHeader.BackupVersion;
+    m_fileHeader.MySQLVersion = 0;
+  }
+  
   debug << "FileHeader: " << m_fileHeader.Magic << " " <<
-    m_fileHeader.NdbVersion << " " <<
+    m_fileHeader.BackupVersion << " " <<
     m_fileHeader.SectionType << " " <<
     m_fileHeader.SectionLength << " " <<
     m_fileHeader.FileType << " " <<
@@ -1003,13 +1521,19 @@ bool RestoreDataIterator::readFragmentHeader(int & ret, Uint32 *fragmentId)
     if (Header.SectionType == BackupFormat::EMPTY_ENTRY)
     {
       void *tmp;
-      buffer_get_ptr(&tmp, Header.SectionLength*4-8, 1);
+      if (Header.SectionLength < 2)
+      {
+        err << "getFragmentFooter:Error reading fragment footer" << endl;
+        return false;
+      }
+      if (Header.SectionLength > 2)
+        buffer_get_ptr(&tmp, Header.SectionLength*4-8, 1);
       continue;
     }
     break;
   }
   /* read rest of header */
-  if (buffer_read(((char*)&Header)+8, sizeof(Header)-8, 1) != 1)
+  if (buffer_read(((char*)&Header)+8, Header.SectionLength*4-8, 1) != 1)
   {
     ret = 0;
     return false;
@@ -1036,6 +1560,8 @@ bool RestoreDataIterator::readFragmentHeader(int & ret, Uint32 *fragmentId)
     return false;
   }
 
+  init_bitfield_storage(m_currentTable->m_dictTable);
+
   info.setLevel(254);
   info << "_____________________________________________________" << endl
        << "Processing data in table: " << m_currentTable->getTableName() 
@@ -1072,7 +1598,7 @@ RestoreDataIterator::validateFragmentFooter() {
 } // RestoreDataIterator::getFragmentFooter
 
 AttributeDesc::AttributeDesc(NdbDictionary::Column *c)
-  : m_column(c)
+  : m_column(c), truncation_detected(false)
 {
   size = 8*NdbColumnImpl::getImpl(* c).m_attrSize;
   arraySize = NdbColumnImpl::getImpl(* c).m_arraySize;
@@ -1086,10 +1612,13 @@ void TableS::createAttr(NdbDictionary::Column *column)
     abort();
   }
   d->attrId = allAttributesDesc.size();
+  d->convertFunc = NULL;
+  d->parameter = NULL;
+  d->m_exclude = false;
   allAttributesDesc.push_back(d);
 
   if (d->m_column->getAutoIncrement())
-    m_auto_val_id= d->attrId;
+    m_auto_val_attrib = d;
 
   if(d->m_column->getPrimaryKey() && backupVersion <= MAKE_VERSION(4,1,7))
   {
@@ -1105,7 +1634,7 @@ void TableS::createAttr(NdbDictionary::Column *column)
   }
 
   // just a reminder - does not solve backwards compat
-  if (backupVersion < MAKE_VERSION(5,1,3))
+  if (backupVersion < MAKE_VERSION(5,1,3) || isDrop6(backupVersion))
   {
     d->m_nullBitIndex = m_noOfNullable; 
     m_noOfNullable++;
@@ -1114,6 +1643,35 @@ void TableS::createAttr(NdbDictionary::Column *column)
   m_variableAttribs.push_back(d);
 } // TableS::createAttr
 
+bool
+TableS::get_auto_data(const TupleS & tuple, Uint32 * syskey, Uint64 * nextid) const
+{
+  /*
+    Read current (highest) auto_increment value for
+    a table. Currently there can only be one per table.
+    The values are stored in sustable SYSTAB_0 as
+    {SYSKEY,NEXTID} values where SYSKEY (32-bit) is
+    the table_id and NEXTID (64-bit) is the next auto_increment
+    value in the sequence (note though that sequences of
+    values can have been fetched and that are cached in NdbAPI).
+    SYSTAB_0 can contain other data so we need to check that
+    the found SYSKEY value is a valid table_id (< 0x10000000).
+   */
+  AttributeData * attr_data = tuple.getData(0);
+  AttributeDesc * attr_desc = tuple.getDesc(0);
+  const AttributeS attr1 = {attr_desc, *attr_data};
+  memcpy(syskey ,attr1.Data.u_int32_value, sizeof(Uint32));
+  attr_data = tuple.getData(1);
+  attr_desc = tuple.getDesc(1);
+  const AttributeS attr2 = {attr_desc, *attr_data};
+  memcpy(nextid, attr2.Data.u_int64_value, sizeof(Uint64));
+  if (*syskey < 0x10000000)
+  {
+    return true;
+  }
+  return false;
+};
+
 Uint16 Twiddle16(Uint16 in)
 {
   Uint16 retVal = 0;
@@ -1193,7 +1751,8 @@ RestoreLogIterator::getNextLogEntry(int & res) {
       return 0;
     }
 
-    if (unlikely(m_metaData.getFileHeader().NdbVersion < NDBD_FRAGID_VERSION))
+    if (unlikely(m_metaData.getFileHeader().NdbVersion < NDBD_FRAGID_VERSION ||
+                 isDrop6(m_metaData.getFileHeader().NdbVersion)))
     {
       /*
         FragId was introduced in LogEntry in version
@@ -1252,10 +1811,10 @@ RestoreLogIterator::getNextLogEntry(int & res) {
 
   AttributeHeader * ah = (AttributeHeader *)attr_data;
   AttributeHeader *end = (AttributeHeader *)(attr_data + attr_data_len);
-  AttributeS *  attr;
+  AttributeS * attr;
   m_logEntry.m_frag_id = frag_id;
   while(ah < end){
-    attr= m_logEntry.add_attr();
+    attr = m_logEntry.add_attr();
     if(attr == NULL) {
       ndbout_c("Restore: Failed to allocate memory");
       res = -1;
@@ -1265,7 +1824,7 @@ RestoreLogIterator::getNextLogEntry(int & res) {
     if(unlikely(!m_hostByteOrder))
       *(Uint32*)ah = Twiddle32(*(Uint32*)ah);
 
-    attr->Desc = (* tab)[ah->getAttributeId()];
+    attr->Desc = tab->getAttributeDesc(ah->getAttributeId());
     assert(attr->Desc != 0);
 
     const Uint32 sz = ah->getDataSize();
@@ -1275,9 +1834,9 @@ RestoreLogIterator::getNextLogEntry(int & res) {
     } else {
       attr->Data.null = false;
       attr->Data.void_value = ah->getDataPtr();
+      Twiddle(attr->Desc, &(attr->Data));
     }
     
-    Twiddle(attr->Desc, &(attr->Data));
     
     ah = ah->getNext();
   }
@@ -1300,7 +1859,9 @@ operator<<(NdbOut& ndbout, const AttributeS& attr){
   
   NdbRecAttr tmprec(0);
   tmprec.setup(desc.m_column, 0);
-  tmprec.receive_data((Uint32*)data.void_value, data.size);
+  Uint32 length = (desc.size)/8 * (desc.arraySize);
+  tmprec.receive_data((Uint32*)data.void_value, length);
+
   ndbrecattr_print_formatted(ndbout, tmprec, g_ndbrecord_print_format);
 
   return ndbout;
@@ -1315,7 +1876,7 @@ operator<<(NdbOut& ndbout, const TupleS& tuple)
     if (i > 0)
       ndbout << g_ndbrecord_print_format.fields_terminated_by;
     AttributeData * attr_data = tuple.getData(i);
-    const AttributeDesc * attr_desc = tuple.getDesc(i);
+    AttributeDesc * attr_desc = tuple.getDesc(i);
     const AttributeS attr = {attr_desc, *attr_data};
     debug << i << " " << attr_desc->m_column->getName();
     ndbout << attr;
