@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,11 +12,13 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <ndb_opts.h>
 #include <Vector.hpp>
+#include <Properties.hpp>
 #include <ndb_limits.h>
 #include <NdbTCP.h>
 #include <NdbMem.h>
@@ -30,11 +33,14 @@ extern FilteredNdbOut err;
 extern FilteredNdbOut info;
 extern FilteredNdbOut debug;
 
+static Uint32 g_tableCompabilityMask = 0;
 static int ga_nodeId = 0;
 static int ga_nParallelism = 128;
 static int ga_backupId = 0;
-static bool ga_dont_ignore_systab_0 = false;
+bool ga_dont_ignore_systab_0 = false;
 static bool ga_no_upgrade = false;
+static bool ga_promote_attributes = false;
+static bool ga_demote_attributes = false;
 static Vector<class BackupConsumer *> g_consumers;
 static BackupPrinter* g_printer = NULL;
 
@@ -54,7 +60,24 @@ unsigned int opt_progress_frequency;
 unsigned int g_report_next;
 Vector<BaseString> g_databases;
 Vector<BaseString> g_tables;
+Vector<BaseString> g_include_tables, g_exclude_tables;
+Vector<BaseString> g_include_databases, g_exclude_databases;
+Properties g_rewrite_databases;
 NdbRecordPrintFormat g_ndbrecord_print_format;
+unsigned int opt_no_binlog;
+
+class RestoreOption
+{
+public:
+  virtual ~RestoreOption() { }
+  int optid;
+  BaseString argument;
+};
+
+Vector<class RestoreOption *> g_include_exclude;
+static void save_include_exclude(int optid, char * argument);
+
+static inline void parse_rewrite_database(char * argument);
 
 NDB_STD_OPTS_VARS;
 
@@ -65,6 +88,7 @@ static bool ga_restore_epoch = false;
 static bool ga_restore = false;
 static bool ga_print = false;
 static bool ga_skip_table_check = false;
+static bool ga_exclude_missing_columns = false;
 static int _print = 0;
 static int _print_meta = 0;
 static int _print_data = 0;
@@ -72,6 +96,11 @@ static int _print_log = 0;
 static int _restore_data = 0;
 static int _restore_meta = 0;
 static int _no_restore_disk = 0;
+static bool _preserve_trailing_spaces = false;
+static bool ga_disable_indexes = false;
+static bool ga_rebuild_indexes = false;
+bool ga_skip_unknown_objects = false;
+bool ga_skip_broken_objects = false;
 BaseString g_options("ndb_restore");
 
 const char *load_default_groups[]= { "mysql_cluster","ndb_restore",0 };
@@ -89,7 +118,16 @@ enum ndb_restore_options {
   OPT_LINES_TERMINATED_BY,
   OPT_APPEND,
   OPT_PROGRESS_FREQUENCY,
-  OPT_VERBOSE
+  OPT_NO_BINLOG,
+  OPT_VERBOSE,
+  OPT_INCLUDE_TABLES,
+  OPT_EXCLUDE_TABLES,
+  OPT_INCLUDE_DATABASES,
+  OPT_EXCLUDE_DATABASES,
+  OPT_REWRITE_DATABASE,
+  OPT_EXCLUDE_MISSING_COLUMNS,
+  OPT_DISABLE_INDEXES,
+  OPT_REBUILD_INDEXES
 };
 static const char *opt_fields_enclosed_by= NULL;
 static const char *opt_fields_terminated_by= NULL;
@@ -98,6 +136,11 @@ static const char *opt_lines_terminated_by= NULL;
 
 static const char *tab_path= NULL;
 static int opt_append;
+static const char *opt_exclude_tables= NULL;
+static const char *opt_include_tables= NULL;
+static const char *opt_exclude_databases= NULL;
+static const char *opt_include_databases= NULL;
+static const char *opt_rewrite_database= NULL;
 
 static struct my_option my_long_options[] =
 {
@@ -123,6 +166,19 @@ static struct my_option my_long_options[] =
     "Don't upgrade array type for var attributes, which don't resize VAR data and don't change column attributes",
     (uchar**) &ga_no_upgrade, (uchar**) &ga_no_upgrade, 0,
     GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "promote-attributes", 'A',
+    "Allow attributes to be promoted when restoring data from backup",
+    (uchar**) &ga_promote_attributes, (uchar**) &ga_promote_attributes, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "lossy-conversions", 'L',
+    "Allow lossy conversions for attributes (type demotions or integral"
+    " signed/unsigned type changes) when restoring data from backup",
+    (uchar**) &ga_demote_attributes, (uchar**) &ga_demote_attributes, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "preserve-trailing-spaces", 'P',
+    "Allow to preserve the tailing spaces (including paddings) When char->varchar or binary->varbinary is promoted",
+    (uchar**) &_preserve_trailing_spaces, (uchar**)_preserve_trailing_spaces , 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
   { "no-restore-disk-objects", 'd',
     "Dont restore disk objects (tablespace/logfilegroups etc)",
     (uchar**) &_no_restore_disk, (uchar**) &_no_restore_disk,  0,
@@ -157,7 +213,7 @@ static struct my_option my_long_options[] =
     (uchar**) &ga_backupPath, (uchar**) &ga_backupPath, 0,
     GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
   { "dont_ignore_systab_0", 'f',
-    "Experimental. Do not ignore system table during restore.", 
+    "Do not ignore system table during --print-data.", 
     (uchar**) &ga_dont_ignore_systab_0, (uchar**) &ga_dont_ignore_systab_0, 0,
     GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
   { "ndb-nodegroup-map", OPT_NDB_NODEGROUP_MAP,
@@ -197,10 +253,57 @@ static struct my_option my_long_options[] =
     "Print status uf restore periodically in given seconds", 
     (uchar**) &opt_progress_frequency, (uchar**) &opt_progress_frequency, 0,
     GET_INT, REQUIRED_ARG, 0, 0, 65535, 0, 0, 0 },
+  { "no-binlog", OPT_NO_BINLOG,
+    "If a mysqld is connected and has binary log, do not log the restored data", 
+    (uchar**) &opt_no_binlog, (uchar**) &opt_no_binlog, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
   { "verbose", OPT_VERBOSE,
     "verbosity", 
     (uchar**) &opt_verbose, (uchar**) &opt_verbose, 0,
     GET_INT, REQUIRED_ARG, 1, 0, 255, 0, 0, 0 },
+  { "include-databases", OPT_INCLUDE_DATABASES,
+    "Comma separated list of databases to restore. Example: db1,db3",
+    (uchar**) &opt_include_databases, (uchar**) &opt_include_databases, 0,
+    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
+  { "exclude-databases", OPT_EXCLUDE_DATABASES,
+    "Comma separated list of databases to not restore. Example: db1,db3",
+    (uchar**) &opt_exclude_databases, (uchar**) &opt_exclude_databases, 0,
+    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
+  { "rewrite-database", OPT_REWRITE_DATABASE,
+    "A pair 'source,dest' of database names from/into which to restore. "
+    "Example: --rewrite-database=oldDb,newDb",
+    (uchar**) &opt_rewrite_database, (uchar**) &opt_rewrite_database, 0,
+    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
+  { "include-tables", OPT_INCLUDE_TABLES, "Comma separated list of tables to "
+    "restore. Table name should include database name. Example: db1.t1,db3.t1", 
+    (uchar**) &opt_include_tables, (uchar**) &opt_include_tables, 0,
+    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
+  { "exclude-tables", OPT_EXCLUDE_TABLES, "Comma separated list of tables to "
+    "not restore. Table name should include database name. "
+    "Example: db1.t1,db3.t1",
+    (uchar**) &opt_exclude_tables, (uchar**) &opt_exclude_tables, 0,
+    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
+  { "exclude-missing-columns", OPT_EXCLUDE_MISSING_COLUMNS,
+    "Ignore columns present in backup but not in database",
+    (uchar**) &ga_exclude_missing_columns,
+    (uchar**) &ga_exclude_missing_columns, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "disable-indexes", OPT_DISABLE_INDEXES,
+    "Disable indexes",
+    (uchar**) &ga_disable_indexes,
+    (uchar**) &ga_disable_indexes, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "rebuild-indexes", OPT_REBUILD_INDEXES,
+    "Rebuild indexes",
+    (uchar**) &ga_rebuild_indexes,
+    (uchar**) &ga_rebuild_indexes, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "skip-unknown-objects", 256, "Skip unknown object when parsing backup",
+    (uchar**) &ga_skip_unknown_objects, (uchar**) &ga_skip_unknown_objects, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
+  { "skip-broken-objects", 256, "Skip broken object when parsing backup",
+    (uchar**) &ga_skip_broken_objects, (uchar**) &ga_skip_broken_objects, 0,
+    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0 },
   { 0, 0, 0, 0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0}
 };
 
@@ -371,13 +474,83 @@ get_one_option(int optid, const struct my_option *opt __attribute__((unused)),
       exit(NDBT_ProgramExit(NDBT_WRONGARGS));
     }
     break;
+  case OPT_INCLUDE_DATABASES:
+  case OPT_EXCLUDE_DATABASES:
+  case OPT_INCLUDE_TABLES:
+  case OPT_EXCLUDE_TABLES:
+    save_include_exclude(optid, argument);
+    break;
+  case OPT_REWRITE_DATABASE:
+    parse_rewrite_database(argument);
+    break;
   }
   return 0;
 }
+
+static const char* SCHEMA_NAME="/def/";
+static const int SCHEMA_NAME_SIZE= 5;
+
+int
+makeInternalTableName(const BaseString &externalName, 
+                      BaseString& internalName)
+{
+  // Make dbname.table1 into dbname/def/table1
+  Vector<BaseString> parts;
+
+  // Must contain a dot
+  if (externalName.indexOf('.') == -1)
+    return -1;
+  externalName.split(parts,".");
+  // .. and only 1 dot
+  if (parts.size() != 2)
+    return -1;
+  internalName.clear();
+  internalName.append(parts[0]); // db name
+  internalName.append(SCHEMA_NAME); // /def/
+  internalName.append(parts[1]); // table name
+  return 0;
+}
+
+void
+processTableList(const char* str, Vector<BaseString> &lst)
+{
+  // Process tables list like db1.t1,db2.t1 and exits when
+  // it finds problems.
+  unsigned int i;
+  /* Split passed string on comma into 2 BaseStrings in the vector */
+  BaseString(str).split(lst,",");
+  for (i=0; i < lst.size(); i++)
+  {
+    BaseString internalName;
+    if (makeInternalTableName(lst[i], internalName))
+    {
+      info << "`" << lst[i] << "` is not a valid tablename!" << endl;
+      exit(NDBT_ProgramExit(NDBT_WRONGARGS));
+    }
+    lst[i].assign(internalName);
+  }
+}
+
+BaseString
+makeExternalTableName(const BaseString &internalName)
+{
+   // Make dbname/def/table1 into dbname.table1
+  int idx;
+  BaseString externalName;
+  
+  idx = internalName.indexOf('/');
+  externalName = internalName.substr(0,idx);
+  externalName.append(".");
+  externalName.append(internalName.substr(idx + SCHEMA_NAME_SIZE,
+                                          internalName.length()));
+  return externalName;
+}
+
 bool
 readArguments(int *pargc, char*** pargv) 
 {
   Uint32 i;
+  BaseString tmp;
   debug << "Load defaults" << endl;
   const char *load_default_groups[]= { "mysql_cluster","ndb_restore",0 };
 
@@ -477,11 +650,26 @@ o verify nodegroup mapping
      restore->m_no_upgrade = true;
   }
 
+  if (_preserve_trailing_spaces)
+  {
+     restore->m_preserve_trailing_spaces = true;
+  }
+
   if (ga_restore_epoch)
   {
     restore->m_restore_epoch = true;
   }
 
+  if (ga_disable_indexes)
+  {
+    restore->m_disable_indexes = true;
+  }
+
+  if (ga_rebuild_indexes)
+  {
+    restore->m_rebuild_indexes = true;
+  }
+
   {
     BackupConsumer * c = g_printer;
     g_consumers.push_back(c);
@@ -513,9 +701,13 @@ o verify nodegroup mapping
   info << "backup path = " << ga_backupPath << endl;
   if (g_databases.size() > 0)
   {
+    info << "WARNING! Using deprecated syntax for selective object restoration." << endl;
+    info << "Please use --include-*/--exclude-* options in future." << endl;
     info << "Restoring only from database " << g_databases[0].c_str() << endl;
     if (g_tables.size() > 0)
-      info << "Restoring only tables:";
+    {
+        info << "Restoring tables:";
+    }
     for (unsigned i= 0; i < g_tables.size(); i++)
     {
       info << " " << g_tables[i].c_str();
@@ -523,6 +715,67 @@ o verify nodegroup mapping
     if (g_tables.size() > 0)
       info << endl;
   }
+  
+  if (opt_include_databases)
+  {
+    tmp = BaseString(opt_include_databases);
+    tmp.split(g_include_databases,",");
+    info << "Including Databases: ";
+    for (i= 0; i < g_include_databases.size(); i++)
+    {
+      info << g_include_databases[i] << " ";
+    }
+    info << endl;
+  }
+  
+  if (opt_exclude_databases)
+  {
+    tmp = BaseString(opt_exclude_databases);
+    tmp.split(g_exclude_databases,",");
+    info << "Excluding databases: ";
+    for (i= 0; i < g_exclude_databases.size(); i++)
+    {
+      info << g_exclude_databases[i] << " ";
+    }
+    info << endl;
+  }
+  
+  if (opt_rewrite_database)
+  {
+    info << "Rewriting databases:";
+    Properties::Iterator it(&g_rewrite_databases);
+    const char * src;
+    for (src = it.first(); src != NULL; src = it.next()) {
+      const char * dst = NULL;
+      bool r = g_rewrite_databases.get(src, &dst);
+      assert(r && (dst != NULL));
+      info << " (" << src << "->" << dst << ")";
+    }
+    info << endl;
+  }
+
+  if (opt_include_tables)
+  {
+    processTableList(opt_include_tables, g_include_tables);
+    info << "Including tables: ";
+    for (i= 0; i < g_include_tables.size(); i++)
+    {
+      info << makeExternalTableName(g_include_tables[i]).c_str() << " ";
+    }
+    info << endl;
+  }
+  
+  if (opt_exclude_tables)
+  {
+    processTableList(opt_exclude_tables, g_exclude_tables);
+    info << "Excluding tables: ";
+    for (i= 0; i < g_exclude_tables.size(); i++)
+    {
+      info << makeExternalTableName(g_exclude_tables[i]).c_str() << " ";
+    }
+    info << endl;
+  }
+  
   /*
     the below formatting follows the formatting from mysqldump
     do not change unless to adopt to changes in mysqldump
@@ -541,6 +794,27 @@ o verify nodegroup mapping
     g_ndbrecord_print_format.null_string= "";
   g_ndbrecord_print_format.hex_prefix= "";
   g_ndbrecord_print_format.hex_format= opt_hex_format;
+
+  if (ga_skip_table_check)
+  {
+    g_tableCompabilityMask = ~(Uint32)0;
+    ga_skip_unknown_objects = true;
+  }
+
+  if (ga_promote_attributes)
+  {
+    g_tableCompabilityMask |= TCM_ATTRIBUTE_PROMOTION;
+  }
+
+  if (ga_demote_attributes)
+  {
+    g_tableCompabilityMask |= TCM_ATTRIBUTE_DEMOTION;
+  }
+
+  if (ga_exclude_missing_columns)
+  {
+    g_tableCompabilityMask |= TCM_EXCLUDE_MISSING_COLUMNS;
+  }
   return true;
 }
 
@@ -555,7 +829,7 @@ clearConsumers()
 static inline bool
 checkSysTable(const TableS* table)
 {
-  return ga_dont_ignore_systab_0 || ! table->getSysTable();
+  return ! table->getSysTable();
 }
 
 static inline bool
@@ -579,15 +853,25 @@ isIndex(const TableS* table)
 }
 
 static inline bool
-checkDbAndTableName(const TableS* table)
+isSYSTAB_0(const TableS* table)
 {
-  if (g_tables.size() == 0 &&
-      g_databases.size() == 0)
-    return true;
-  if (g_databases.size() == 0)
-    g_databases.push_back("TEST_DB");
+  return table->isSYSTAB_0();
+}
 
-  // Filter on the main table name for indexes and blobs
+static inline bool
+isInList(BaseString &needle, Vector<BaseString> &lst){
+  unsigned int i= 0;
+  for (i= 0; i < lst.size(); i++)
+  {
+    if (strcmp(needle.c_str(), lst[i].c_str()) == 0)
+      return true;
+  }
+  return false;
+}
+
+const char*
+getTableName(const TableS* table)
+{
   const char *table_name;
   if (isBlobTable(table))
     table_name= table->getMainTable()->getTableName();
@@ -596,6 +880,169 @@ checkDbAndTableName(const TableS* table)
       NdbTableImpl::getImpl(*table->m_dictTable).m_primaryTable.c_str();
   else
     table_name= table->getTableName();
+    
+  return table_name;
+}
+
+static void parse_rewrite_database(char * argument)
+{
+  const BaseString arg(argument);
+  Vector<BaseString> args;
+  unsigned int n = arg.split(args, ",");
+  if ((n == 2)
+      && (args[0].length() > 0)
+      && (args[1].length() > 0)) {
+    const BaseString src = args[0];
+    const BaseString dst = args[1];
+    const bool replace = true;
+    bool r = g_rewrite_databases.put(src.c_str(), dst.c_str(), replace);
+    assert(r);
+    return; // ok
+  }
+
+  info << "argument `" << arg.c_str()
+       << "` is not a pair 'a,b' of non-empty names." << endl;
+  exit(NDBT_ProgramExit(NDBT_WRONGARGS));
+}
+
+static void save_include_exclude(int optid, char * argument)
+{
+  BaseString arg = argument;
+  Vector<BaseString> args;
+  arg.split(args, ",");
+  for (uint i = 0; i < args.size(); i++)
+  {
+    RestoreOption * option = new RestoreOption();
+    BaseString arg;
+    
+    option->optid = optid;
+    switch (optid) {
+    case OPT_INCLUDE_TABLES:
+    case OPT_EXCLUDE_TABLES:
+      if (makeInternalTableName(args[i], arg))
+      {
+        info << "`" << args[i] << "` is not a valid tablename!" << endl;
+        exit(NDBT_ProgramExit(NDBT_WRONGARGS));
+      }
+      break;
+    default:
+      arg = args[i];
+      break;
+    }
+    option->argument = arg;
+    g_include_exclude.push_back(option);
+  }
+}
+static bool check_include_exclude(BaseString database, BaseString table)
+{
+  const char * db = database.c_str();
+  const char * tbl = table.c_str();
+  bool do_include = true;
+
+  if (g_include_databases.size() != 0 ||
+      g_include_tables.size() != 0)
+  {
+    /*
+      User has explicitly specified what databases
+      and/or tables should be restored. If no match is
+      found then DON'T restore table.
+     */
+    do_include = false;
+  }
+  if (do_include &&
+      (g_exclude_databases.size() != 0 ||
+       g_exclude_tables.size() != 0))
+  {
+    /*
+      User has not explicitly specified what databases
+      and/or tables should be restored.
+      User has explicitly specified what databases
+      and/or tables should NOT be restored. If no match is
+      found then DO restore table.
+     */
+    do_include = true;
+  }
+
+  if (g_include_exclude.size() != 0)
+  {
+    /*
+      Scan include exclude arguments in reverse.
+      First matching include causes table to be restored.
+      first matching exclude causes table NOT to be restored.      
+     */
+    for(uint i = g_include_exclude.size(); i > 0; i--)
+    {
+      RestoreOption *option = g_include_exclude[i-1];
+      switch (option->optid) {
+      case OPT_INCLUDE_TABLES:
+        if (strcmp(tbl, option->argument.c_str()) == 0)
+          return true; // do include
+        break;
+      case OPT_EXCLUDE_TABLES:
+        if (strcmp(tbl, option->argument.c_str()) == 0)
+          return false; // don't include
+        break;
+      case OPT_INCLUDE_DATABASES:
+        if (strcmp(db, option->argument.c_str()) == 0)
+          return true; // do include
+        break;
+      case OPT_EXCLUDE_DATABASES:
+        if (strcmp(db, option->argument.c_str()) == 0)
+          return false; // don't include
+        break;
+      default:
+        continue;
+      }
+    }
+  }
+  
+  return do_include;
+}
+
+static inline bool
+checkDoRestore(const TableS* table)
+{
+  int idx;
+  bool ret = true;
+  BaseString db, tbl;
+  
+  tbl.assign(getTableName(table));
+  idx = tbl.indexOf('/');
+  db = tbl.substr(0, idx);
+  
+  /*
+    Include/exclude flags are evaluated right
+    to left, and first match overrides any other
+    matches. Non-overlapping arguments are accumulative.
+    If no include flags are specified this means all databases/tables
+    except any excluded are restored.
+    If include flags are specified than only those databases
+    or tables specified are restored.
+   */
+  ret = check_include_exclude(db, tbl);
+  return ret;
+}
+
+static inline bool
+checkDbAndTableName(const TableS* table)
+{
+  if (table->isBroken())
+    return false;
+
+  // If new options are given, ignore the old format
+  if (opt_include_tables || opt_exclude_tables ||
+      opt_include_databases || opt_exclude_databases ) {
+    return (checkDoRestore(table));
+  }
+  
+  if (g_tables.size() == 0 && g_databases.size() == 0)
+    return true;
+
+  if (g_databases.size() == 0)
+    g_databases.push_back("TEST_DB");
+
+  // Filter on the main table name for indexes and blobs
+  const char *table_name= getTableName(table);
 
   unsigned i;
   for (i= 0; i < g_databases.size(); i++)
@@ -618,13 +1065,11 @@ checkDbAndTableName(const TableS* table)
   while (*table_name != '/') table_name++;
   table_name++;
 
+  // Check if table should be restored
   for (i= 0; i < g_tables.size(); i++)
   {
     if (strcmp(table_name, g_tables[i].c_str()) == 0)
-    {
-      // we have a match
       return true;
-    }
   }
   return false;
 }
@@ -648,20 +1093,20 @@ static void exitHandler(int code)
 
 static void init_progress()
 {
-  struct timeval the_time;
-  gettimeofday(&the_time, 0);
-  g_report_next = the_time.tv_sec + opt_progress_frequency;
+  Uint64 now = NdbTick_CurrentMillisecond() / 1000;
+  g_report_next = now + opt_progress_frequency;
 }
 
 static int check_progress()
 {
   if (!opt_progress_frequency)
     return 0;
-  struct timeval the_time;
-  gettimeofday(&the_time, 0);
-  if (the_time.tv_sec >= g_report_next)
+
+  Uint64 now = NdbTick_CurrentMillisecond() / 1000;
+  
+  if (now  >= g_report_next)
   {
-    g_report_next = the_time.tv_sec + opt_progress_frequency;
+    g_report_next = now + opt_progress_frequency;
     return 1;
   }
   return 0;
@@ -677,6 +1122,27 @@ static void report_progress(const char *prefix, const BackupFile &f)
     info << prefix << f.get_file_pos() << " bytes\n";
 }
 
+/**
+ * Reports, clears information on columns where data truncation was detected.
+ */
+static void
+check_data_truncations(const TableS * table)
+{
+  assert(table);
+  const char * tname = table->getTableName();
+  const int n = table->getNoOfAttributes();
+  for (int i = 0; i < n; i++) {
+    AttributeDesc * desc = table->getAttributeDesc(i);
+    if (desc->truncation_detected) {
+      const char * cname = desc->m_column->getName();
+      info.setLevel(254);
+      info << "Data truncation(s) detected for attribute: "
+           << tname << "." << cname << endl;
+      desc->truncation_detected = false;
+    }
+  }
+}
+
 int
 main(int argc, char** argv)
 {
@@ -688,12 +1154,18 @@ main(int argc, char** argv)
     exitHandler(NDBT_FAILED);
   }
 
-  g_options.appfmt(" -b %d", ga_backupId);
+  g_options.appfmt(" -b %u", ga_backupId);
   g_options.appfmt(" -n %d", ga_nodeId);
   if (_restore_meta)
     g_options.appfmt(" -m");
   if (ga_no_upgrade)
     g_options.appfmt(" -u");
+  if (ga_promote_attributes)
+    g_options.appfmt(" -A");
+  if (ga_demote_attributes)
+    g_options.appfmt(" -L");
+  if (_preserve_trailing_spaces)
+    g_options.appfmt(" -P");
   if (ga_skip_table_check)
     g_options.appfmt(" -s");
   if (_restore_data)
@@ -702,7 +1174,17 @@ main(int argc, char** argv)
     g_options.appfmt(" -e");
   if (_no_restore_disk)
     g_options.appfmt(" -d");
+  if (ga_exclude_missing_columns)
+    g_options.append(" --exclude-missing-columns");
+  if (ga_disable_indexes)
+    g_options.append(" --disable-indexes");
+  if (ga_rebuild_indexes)
+    g_options.append(" --rebuild-indexes");
   g_options.appfmt(" -p %d", ga_nParallelism);
+  if (ga_skip_unknown_objects)
+    g_options.append(" --skip-unknown-objects");
+  if (ga_skip_broken_objects)
+    g_options.append(" --skip-broken-objects");
 
   g_connect_string = opt_connect_str;
 
@@ -720,12 +1202,22 @@ main(int argc, char** argv)
   }
 
   const BackupFormat::FileHeader & tmp = metaData.getFileHeader();
-  const Uint32 version = tmp.NdbVersion;
+  const Uint32 version = tmp.BackupVersion;
   
   char buf[NDB_VERSION_STRING_BUF_SZ];
   info.setLevel(254);
-  info << "Ndb version in backup files: " 
-       <<  ndbGetVersionString(version, 0, 0, buf, sizeof(buf)) << endl;
+  info << "Backup version in files: " 
+       <<  ndbGetVersionString(version, 0, 
+                               isDrop6(version) ? "-drop6" : 0, 
+                               buf, sizeof(buf));
+  if (version >= NDBD_RAW_LCP)
+  {
+    info << " ndb version: "
+         << ndbGetVersionString(tmp.NdbVersion, tmp.MySQLVersion, 0, 
+                                buf, sizeof(buf));
+  }
+
+  info << endl;
   
   /**
    * check wheater we can restore the backup (right version).
@@ -774,7 +1266,7 @@ main(int argc, char** argv)
   Uint32 i;
   for(i= 0; i < g_consumers.size(); i++)
   {
-    if (!g_consumers[i]->init())
+    if (!g_consumers[i]->init(g_tableCompabilityMask))
     {
       clearConsumers();
       err << "Failed to initialize consumers" << endl;
@@ -815,6 +1307,10 @@ main(int argc, char** argv)
     table_output.push_back(NULL);
     if (!checkDbAndTableName(table))
       continue;
+    if (isSYSTAB_0(table))
+    {
+      table_output[i]= ndbout.m_out;
+    }
     if (checkSysTable(table))
     {
       if (!tab_path || isBlobTable(table) || isIndex(table))
@@ -885,18 +1381,20 @@ main(int argc, char** argv)
   {
     if(_restore_data || _print_data)
     {
-      if (!ga_skip_table_check){
-        for(i=0; i < metaData.getNoOfTables(); i++){
-          if (checkSysTable(metaData, i))
+      // Check table compatibility
+      for (i=0; i < metaData.getNoOfTables(); i++){
+        if (checkSysTable(metaData, i) &&
+            checkDbAndTableName(metaData[i]))
+        {
+          for(Uint32 j= 0; j < g_consumers.size(); j++)
           {
-            for(Uint32 j= 0; j < g_consumers.size(); j++)
-              if (!g_consumers[j]->table_equal(* metaData[i]))
-              {
-                err << "Restore: Failed to restore data, ";
-                err << metaData[i]->getTableName() << " table structure doesn't match backup ... Exiting " << endl;
-                exitHandler(NDBT_FAILED);
-              }
-          }
+            if (!g_consumers[j]->table_compatible_check(*metaData[i]))
+            {
+              err << "Restore: Failed to restore data, ";
+              err << metaData[i]->getTableName() << " table structure incompatible with backup's ... Exiting " << endl;
+              exitHandler(NDBT_FAILED);
+            } 
+          } 
         }
       }
       RestoreDataIterator dataIter(metaData, &free_data_callback);
@@ -1001,9 +1499,10 @@ main(int argc, char** argv)
     
     if(_restore_data)
     {
-      for(i = 0; i<metaData.getNoOfTables(); i++)
+      for(i = 0; i < metaData.getNoOfTables(); i++)
       {
         const TableS* table = metaData[i];
+        check_data_truncations(table);
         OutputStream *output = table_output[table->getLocalId()];
         if (!output)
           continue;
@@ -1037,7 +1536,25 @@ main(int argc, char** argv)
                "please look at configuration.");
     }               
   }
-  
+
+  if (ga_rebuild_indexes)
+  {
+    debug << "Rebuilding indexes" << endl;
+    for(i = 0; i<metaData.getNoOfTables(); i++)
+    {
+      const TableS *table= metaData[i];
+      if (! (checkSysTable(table) && checkDbAndTableName(table)))
+        continue;
+      if (isBlobTable(table) || isIndex(table))
+        continue;
+      for(Uint32 j= 0; j < g_consumers.size(); j++)
+      {
+        if (!g_consumers[j]->rebuild_indexes(* table))
+          return -1;
+      }
+    }
+  }
+
   /* report to clusterlog if applicable */
   for (i = 0; i < g_consumers.size(); i++)
     g_consumers[i]->report_completed(ga_backupId, ga_nodeId);
@@ -1063,3 +1580,4 @@ main(int argc, char** argv)
 
 template class Vector<BackupConsumer*>;
 template class Vector<OutputStream*>;
+template class Vector<RestoreOption *>;
